/* wrProxyBackendLib.c - proxy backend interface library */

/*
 * Copyright (c) 2005-2006 Wind River Systems, Inc.  
 *
 * The right to copy, distribute, modify or otherwise make use  of this 
 * software may be licensed only pursuant to the terms of an applicable Wind 
 * River license agreement.
 */

/*
modification history
--------------------
01b,16jun06,dbt  Added routine to validate backend arguments.
01a,20may05,dbt  written.
*/

/*
DESCRIPTION

This library contains the Target Agent Proxy backend interface
*/

#ifndef	HOST
#include <vxWorks.h>
#else	/* HOST */
#include <host.h>
#endif	/* HOST */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#ifndef	WIN32
#   include <sys/ioctl.h>
#else	/* WIN32 */
#   include <rpc/rpc.h>
#   include <io.h>
#endif	/* WIN32 */
#include <sys/types.h>
#include <fcntl.h>
#include "wrProxyLibP.h"
#include "wrProxyBackendLib.h"
#ifdef	HOST
#include "wrProxyWdbSerialProtLib.h"
#endif	/* HOST */

/* local variables */

static PROXY_BACKEND *	pBackendList = NULL;	/* backend list */

/* static declarations */

static PROXY_BACKEND *	wrProxyBackendFind (const char * backendName);

/*******************************************************************************
*
* wrProxyBackendFind - find a backend
*
* This routine finds a backend in the registered backend list given its
* name.
*
* RETURNS: pointer on backend structure or NULL if it can not be found
*/

static PROXY_BACKEND * wrProxyBackendFind
    (
    const char *	backendName	/* backend name */
    )
    {
    PROXY_BACKEND *	pBackend = NULL;

    for (pBackend = pBackendList; pBackend != NULL; pBackend = pBackend->next)
    	{
	if (strcmp (pBackend->name, backendName) == 0)
	    break;
	}

    return (pBackend);
    }

/*******************************************************************************
*
* wrProxyBackendRegister - register a specific backend
*
* This routine registers a new backend
*
* RETURNS: OK or ERROR if there was a problem registering the backend
*/

STATUS wrProxyBackendRegister
    (
    const char *		backendName,	/* backend name */
    PROXY_BKEND_OPEN_RTN	openRtn,	/* backend open routine */
    PROXY_BKEND_READ_RTN	readRtn,	/* backend read routine */
    PROXY_BKEND_WRITE_RTN	writeRtn,	/* backend write routine */
    PROXY_BKEND_CLOSE_RTN	closeRtn	/* backend close routine */
    )
    {
    PROXY_BACKEND *	pBackend;

    if (wrProxyBackendFind (backendName) != NULL)
    	{
	printf ("Proxy backend %s is already registered.\n", backendName);
    	return (ERROR);
	}

    if ((pBackend = calloc (1,
    		sizeof (PROXY_BACKEND) + strlen (backendName) + 1)) == NULL)
	{
	printf ("Error registering proxy backend %s (not enough memory).\n",
			backendName);
    	return (ERROR);
	}

    pBackend->next		= (PROXY_BACKEND *) pBackendList;
    pBackend->name		= (char *)pBackend + sizeof (PROXY_BACKEND);
    strcpy (pBackend->name, backendName);
    pBackend->openRtn		= openRtn;
    pBackend->readRtn		= readRtn;
    pBackend->writeRtn		= writeRtn;
    pBackend->closeRtn		= closeRtn;

    /* Update backend list */

    pBackendList		= pBackend;

    return (OK);
    }

/*******************************************************************************
*
* wrProxyBackendConnect - backend connect
*
* This routine hanles the target agent connect request. It looks for the
* correct backend and tries to open a connection with the target agent.
*
* RETURNS: OK or ERROR
*/

STATUS wrProxyBackendConnect
    (
    PROXY_CONNECTION *	pProxy,		/* Proxy connection to connect to */
    char *		args,		/* arguments */
    char *		errorMsg,	/* where to store error message */
    int			errorMsgLength, /* error message max length */
    BOOL		dryRun		/* simulate backend opening to */
    					/* validate arguments */
    )
    {
    char		buffer [100];
    PROXY_BACKEND *	pBackend;
    int			status;

    if (pProxy == NULL && dryRun == FALSE)
    	{
	snprintf (errorMsg, errorMsgLength, "wrproxy: Internal error.");
    	return (ERROR);
	}

    if (wrProxyClientCommandArgGet (args, "type=", NULL, buffer, 100) != 1)
    	{
	strncpy (errorMsg, "wrproxy: Connection type not specified",
			    errorMsgLength);
	errorMsg [errorMsgLength - 1] = EOS;
    	return (ERROR);
	}

    if ((pBackend = wrProxyBackendFind (buffer)) == NULL)
    	{
	snprintf (errorMsg, errorMsgLength,
			"wrproxy: Can not find backend for connection %s.",
			buffer);
	return (ERROR);
	}

    status = wrProxyClientCommandArgGet (args, "protocol=", NULL, buffer, 100);
    
    if (status == -1)
    	{
	strncpy (errorMsg, "wrproxy: Unable to parse protocol parameter",
			    errorMsgLength);
	errorMsg [errorMsgLength - 1] = EOS;
    	return (ERROR);
	}
    else if (status == 1)
        {
	if (strcmp (buffer, "raw") == 0)
	    {
	    if (!dryRun)
	    	pProxy->protocol = PROXY_PROT_RAW;
	    }
	else if (strcmp (buffer, "wdbserial") == 0)
	    {
	    if (!dryRun)
		pProxy->protocol = PROXY_PROT_WDBSERIAL;
	    }
	else
	    {
	    snprintf (errorMsg, errorMsgLength,
			    "wrproxy: Unknown protocol %s.",
			    buffer);
	    return (ERROR);
	    }
	}
    else
    	{
    	if (!dryRun)
	    pProxy->protocol = PROXY_PROT_RAW;
	}

    if (pBackend->openRtn (pProxy, args, errorMsg, errorMsgLength, dryRun) != 0)
	return (ERROR);

    if (dryRun)
    	return (OK);

#ifdef	HOST
    if (pProxy->protocol == PROXY_PROT_WDBSERIAL)
    	{
	/*
	 * Setup buffer that will be used for WDB serial protocol encoding 
	 * and decoding.
	 */

	pProxy->buffer = malloc (PROXY_MAX_PACKET_SIZE);
	if (pProxy->buffer == NULL)
	    {
	    strncpy (errorMsg, "wrproxy: Unable to allocate memory",
				errorMsgLength);
	    errorMsg [errorMsgLength - 1] = EOS;
	    return (ERROR);
	    }
	pProxy->bufIx = 0;
	}
#endif	/* HOST */

    pProxy->pBackend = pBackend;

    return (OK);
    }

/*******************************************************************************
*
* wrProxyBackendWrite - write data to the backend
*
* This routine writes data to the backend. It is also in charge of encoding
* datas if a specific protocol is used to communicate with the target agent.
*
* RETURNS: number of bytes written or ERROR if there was a problem.
*/

int wrProxyBackendWrite
    (
    PROXY_CONNECTION *	pProxy,		/* proxy connection to write to */
    char *		buffer,		/* buffer to write */
    int			nBytes		/* number of bytes to write */
    )
    {
    if (pProxy->pBackend->writeRtn == NULL)
    	return (ERROR);

#ifdef	HOST
    if (pProxy->protocol == PROXY_PROT_WDBSERIAL)
    	{
	/* Clear receive index since we start a new WDB request */

	pProxy->bufIx = 0;

	/* Use buffer in PROXY structure to encode data */

	nBytes = wrProxyWdbSerialProtEncode
			(buffer, nBytes, pProxy->buffer, PROXY_MAX_PACKET_SIZE);

	return (pProxy->pBackend->writeRtn (pProxy, pProxy->buffer, nBytes));
	}
    else
#endif	/* HOST */
	return (pProxy->pBackend->writeRtn (pProxy, buffer, nBytes));
    }

/*******************************************************************************
*
* wrProxyBackendRead - read data from the backend
*
* This routine reads data from the backend. It is also in charge of decoding
* datas if a specific protocol is used to communicate with the target agent.
*
* RETURNS: number of bytes read or ERROR if there was a problem.
*/

int wrProxyBackendRead
    (
    PROXY_CONNECTION *	pProxy,		/* proxy connetion to read from */
    char *		buffer,		/* buffer to put read data */
    int			bufLength	/* buffer length */
    )
    {
    if (pProxy->pBackend->readRtn == NULL)
    	return (ERROR);

#ifdef	HOST
    if (pProxy->protocol == PROXY_PROT_WDBSERIAL)
    	{
	int	nBytes;
	char	inChar;

	/*
	 * Okay, this is not the best way to do but let's read data byte
	 * per byte until we receive the FRAME_END character. We should
	 * improve this in the future.
	 */

	while (TRUE)
	    {
	    /*
	     * Check if buffer is full; if yes, skip packet, it is certainly
	     * not a WDB serial packet
	     */

	    if (pProxy->bufIx >= PROXY_MAX_PACKET_SIZE)
		pProxy->bufIx = 0;

	    /* read one character */

	    if (pProxy->pBackend->readRtn(pProxy, &inChar, 1) != 1)
	    	return (0);

	    if (pProxy->bufIx == 0)
	    	{
		/*
		 * We start storing packet only when we receive frame end/start
		 * character.
		 */

		if (!wrProxyWdbSerialProtIsFrameEnd (inChar))
		     continue;
		}

	    /* Store received char and increment index */

	    pProxy->buffer[pProxy->bufIx++] = inChar;

	    /* Check if we have completely received packet */

	    if (wrProxyWdbSerialProtIsFrameEnd (inChar) && pProxy->bufIx != 1)
	    	{
		/* The packet has been fully received, encode it and return */

		nBytes = wrProxyWdbSerialProtDecode (pProxy->buffer,
						    pProxy->bufIx,
						    buffer, bufLength);

		/* Clear packet receive index */

		pProxy->bufIx = 0;

		return (nBytes);
		}
	    }
	}
    else
#endif	/* HOST */
	return (pProxy->pBackend->readRtn (pProxy,
					    buffer, bufLength));
    }

/*******************************************************************************
*
* wrProxyBackendClose - Close connection with the target agent
*
* This routine closes the connection with the target agent.
*
* RETURNS: N/A
*/

void wrProxyBackendClose
    (
    PROXY_CONNECTION *	pProxy	/* proxy connection used to close backend */
    )
    {
    if (pProxy->pBackend && pProxy->pBackend->closeRtn)
	pProxy->pBackend->closeRtn (pProxy);
#ifdef	HOST
    if (pProxy->buffer != NULL)
	free (pProxy->buffer);
#endif	/* HOST */
    }
