/* wrProxyLib.c - proxy library */

/*
 * Copyright (c) 2005-2006 Wind River Systems, Inc.  
 *
 * The right to copy, distribute, modify or otherwise make use  of this 
 * software may be licensed only pursuant to the terms of an applicable Wind 
 * River license agreement.
 */

/*
modification history
--------------------
01d,20jun06,dbt  Added support for dynamic creation of connections.
		 Added support for init string
		 Misc fixes.
		 Updated version number
01c,28feb06,bpn  Check setsockopt() return value (SPR#117676).
01b,22aug05,dbt  Fixed error handling.
01a,20may05,dbt  written.
*/

/*
DESCRIPTION

This library contains the core of the Wind River Proxy. This proxy is mainly 
used to connect host tools to a WDB agent but it can be used for other
purposes.
*/

/* includes */

#ifndef	HOST
#include <vxWorks.h>
#include <ctype.h>
#else
#include <host.h>
#endif	/* HOST */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#ifdef	HOST
#include <ctype.h>
#include <unistd.h>
#include <sys/time.h>
#ifndef	WIN32
#include <strings.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/tcp.h>
#else	/* WIN32 */
#include <rpc/rpc.h>
#include <io.h>
#endif	/* WIN32 */
#else	/* HOST */
#include <sockLib.h>
#include <selectLib.h>
#include <netinet/tcp.h>
#endif	/* HOST */
#include <netdb.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include "wrProxyLibP.h"
#include "wrProxyBackendLib.h"

/* defines */

#undef	DEBUG

#define	ERROR_MSG_LENGTH	200	/* maximum error message length */
#define	NUMBER_STR_LENGTH	10	/* maximum length of a string used */
					/* to represent a decimal or */
					/* hexadecimal number. */
#define COMMAND_SEPARATOR	'\n'	/* single command separator */

/* error messages */

#define MEM_ALLOC_FAILURE	"wrproxy: Error allocating memory"
#define	UNKNOWN_COMMAND		"wrproxy: Unknown command %s"
#define	MODE_PARSE_ERROR	"wrproxy: Error parsing \"mode\" argument"
#define	UNKNOWN_MODE		"wrproxy: Unknown mode %s"
#define	CMDLINE_PARSE_ERROR	"wrproxy: Error parsing command line"
#define	STARTUP_SCRIPT_ERROR 	"wrproxy: \"%s\" command can not be called " \
				"from startup script"
#define PPORT_PARSE_ERROR	"wrproxy: Unable to parse pport argument"
#define	CONNECTION_NOT_FOUND	"wrproxy: Can not find specified connection"
#define	SOCKET_CREATION_ERROR	"wrproxy: Error creating socket"
#define SOCKET_BIND_ERROR	"wrproxy: Error binding socket"
#define	GETSOCKNAME_ERROR	"wrproxy: getsockname error"
#define	SOCKET_LISTEN_ERROR	"wrproxy: Error listening socket"
#define	SOCKET_ACCEPT_ERROR	"wrproxy: Error accepting socket connection"
#define	PPORT_OPTION_NAME	"pport="
#define PPORT_ALREADY_IN_USE	\
			"wrproxy: Specified proxy port is already in use"
#define	UNAUTHORIZED_COMMAND	"wrproxy: Unauthorized command."

/* typedefs */

typedef struct proxy_cmd
    {
    char *	cmdName;
    STATUS 	(*cmdHandleRtn)
		    (
		    PROXY_CONNECTION *	pProxy,
		    char *		errorMsg,
		    int			errorMsgLength,
		    char **		pResult,
		    char *		args
		    );
    } PROXY_CMD;

typedef struct proxy_server	/* Proxy Server description */
    {
    void *			next;		/* next proxy connection */
    						/* in list */
    void *			prev;		/* previous proxy connection */
    						/* in list */
    int				sockId;		/* server socket id */
    char *			cnctArgs;	/* optional connect args */
    u_short			portNumber;	/* port number used for this */
    						/* server connection */
    int				cnctMaxNum;	/* maximum number of */
    						/* simultaneous connections */
						/* for this server */
    int				cnctNum;	/* current number of */
    						/* connections for this */
						/* server */
    } PROXY_SERVER;

/* local variables */

static char 	errorMsg [ERROR_MSG_LENGTH];	/* error message */
static char 	buffer [PROXY_MAX_PACKET_SIZE];	/* misc. buffer */
static PROXY_SERVER *	pServerList = NULL;	/* proxy server list */
static PROXY_CONNECTION * pProxyList = NULL;	/* proxy connection list */
static void (*cmdHandleCallbackRtn) (char * result); /* proxy command call */
						/* back routine */

/* forward declaration */

static void	wrProxyErrnoPrint		(char * message,
						 char * errorMsg,
						 int errorMsgLength);
static void	wrProxyConnectionOpen		(PROXY_SERVER * pServer);
static void	wrProxyConnectionClose		(PROXY_CONNECTION * pProxy);
static STATUS	wrProxyClientRead		(PROXY_CONNECTION * pProxy);
static int	wrProxyClientWrite		(PROXY_CONNECTION * pProxy,
						 char * buffer, int nBytes);
static int	wrProxyClientCommandParse	(char * buffer,
						 char ** pCommand,
						 char ** pArgs);
static STATUS	wrProxyClientAckSend		(PROXY_CONNECTION *  pProxy,
						 STATUS result, char * data);
static STATUS	wrProxyTgtRead			(PROXY_CONNECTION *  pProxy);
static STATUS	wrProxyConnectCmdHandle		(PROXY_CONNECTION * pProxy,
						 char * errorMsg,
						 int errorMsgLength,
						 char ** pResult,
						 char * args);
static void	wrProxyTcpSocketOptionsSet	(int sockId);
static PROXY_SERVER * wrProxyServerCreate	(u_short portNumber,
						 int cnctMaxNum,
						 char * cnctArgs,
						 char * errorMsg,
						 int errorMsgLength);
static STATUS	wrProxyCommandRun		(PROXY_CONNECTION * pProxy,
						 char * command);
#ifdef	HOST
static STATUS	wrProxyCreateCmdHandle		(PROXY_CONNECTION * pProxy,
						 char * errorMsg,
						 int errorMsgLength,
						 char ** pResult,
						 char * args);
static STATUS	wrProxyDeleteCmdHandle		(PROXY_CONNECTION * pProxy,
						 char * errorMsg,
						 int errorMsgLength,
						 char ** pResult,
						 char * args);
static STATUS	wrProxyListCmdHandle		(PROXY_CONNECTION * pProxy,
						 char * errorMsg,
						 int errorMsgLength,
						 char ** pResult,
						 char * args);
static STATUS	wrProxyInitStringRun		(char * initString);
static void	wrProxyClientCommandArgRemove	(char * args,
						 const char * argName);
static void	wrProxyCmdCallback		(char * result);
#endif	/* HOST */

/* Proxy command array */

static PROXY_CMD proxyCmdArray [] = 
    {
#ifdef	HOST
    {"connect",	wrProxyConnectCmdHandle},
    {"create",	wrProxyCreateCmdHandle},
    {"delete",	wrProxyDeleteCmdHandle},
    {"list",	wrProxyListCmdHandle}
#else	/* !HOST */
    {"connect",	wrProxyConnectCmdHandle}
#endif	/* HOST */
    };

/*******************************************************************************
*
* wrProxyClientCommandArgGet - get a client command argument
*
* This routine gets a command argument given its name. If <pIntResult> is 
* not NULL, the argument is interpreted as an argument and the result is
* stored at the address pointed to by <pIntResult>. If <pCharResult> is not NULL
* the argument is interpreted as a string and the result is stored in the 
* buffer pointed to by <pCharResul>.
*
* RETURNS: 0 if argument was not found, -1 if argument was found but 
*          incorrectly parsed, 1 if argument was found and correctly parsed.
*/

int wrProxyClientCommandArgGet
    (
    char *		args,		/* arguments to parse */
    const char *	argName,	/* argument name */
    int *		pIntResult,	/* where to store integer result */
    char *		pCharResult,	/* where to store string result */
    UINT32		length		/* string result buffer length */
    )
    {
    char *	start;
    char *	end;
    char	savedChar = EOS;
    int		status = -1;
    char *	ptr;

    /* Verify parameters are correctly set */

    if (((pCharResult != NULL) && (pIntResult != NULL)) ||
	(argName == NULL))
    	return (-1);

    /*
     * Verify if the argument list is not empty. If it is indicate we have
     * not found the specified argument.
     */

    if (args == NULL)
    	return (0);

    /* look for argument argName in argument list */
    
    ptr = args;
again:
    if ((start = strstr (ptr, argName)) == NULL)
    	return (0);
    else if ((start != (char *) args) &&
    	     (*(start - 1) != ';') &&
	     (!isspace ((int) *(start - 1))))
    	{
	/* Skip this occurence of the found string and look for another one */

	ptr = start + 1;
	goto again;
	}

    /* Get rid off "<args>=" string */

    start += strlen (argName);

    /* We have found the argument; retrieve its value */

    if ((end = strchr (start, ';')) != NULL)
    	{
	savedChar = *end;
	*end = EOS;
	}

    if (pCharResult != NULL)
    	{
	if ((strlen (start) + 1) > length)
		return (-1);
    	strcpy (pCharResult, start);
	status = 1;
	}
    else if (pIntResult != NULL)
    	{
	char *	ptr;
	int		result;

	result = strtol (start, (char **) &ptr, 0);	/* get the integer */

	/*
	 * A string is correctly evaluated if the entire parameter has
	 * been parsed.
	 */

	if (((result == 0) && (ptr == start)) ||/* wrong evaluation ? */
	    (*ptr != EOS))			/* char remaining ? */
	    status = -1;			/* bad parse! */
	else
	    {
	    *pIntResult = result;
	    status = 1;
	    }
	}
    else
    	status = -1;

    if (savedChar != (char) EOS)
    	*end = savedChar;
    return (status);
    }

/*******************************************************************************
*
* wrProxyTcpSocketOptionsSet - set options for a TCP socket
*
* This routine is used to set default options used by wrproxy for TCP sockets.
*
* RETURNS: N/A
*/

static void wrProxyTcpSocketOptionsSet
    (
    int		sockId		/* socket id */
    )
    {
    struct linger	optVal;
    int			on;

    /*
     * set SO_LINGER & SO_REUSEADDR socket options so that it closes the
     * connections gracefully, when required to close.
     */

    optVal.l_onoff = 1;
    optVal.l_linger = 0;

    if (setsockopt (sockId, SOL_SOCKET, SO_LINGER,
    				(void *)&optVal, sizeof (optVal)) != OK)
	{
	wrProxyErrnoPrint ("Unable to set SO_LINGER socket option: ",
				NULL, 0);
	};

#ifndef	WIN32
    /*
     * Don't use SO_REUSEADDR on Windows, its behavior is different. It allows
     * several sockets to be bound to the same IP address!!!
     */

    on = 1;
    if (setsockopt (sockId, SOL_SOCKET, SO_REUSEADDR,
		(char *)&on, sizeof(on)) != OK)
	{
	wrProxyErrnoPrint ("Unable to set SO_REUSEADDR socket option: ",
				NULL, 0);
	}
#endif

    /* Set TCP_NODELAY socket option to optimize communication */

    on = 1;
    if (setsockopt (sockId, IPPROTO_TCP, TCP_NODELAY,
    		    (void *)&on, sizeof(on)) != OK)
	{
	wrProxyErrnoPrint ("Unable to set socket option to TCP_NODELAY: ",
				NULL, 0);
	}
    }

/*******************************************************************************
*
* wrProxyClientCommandParse - parse a client command
*
* This routine parses a command received from a client. Pointer on the command
* name is stored at the address pointed to by <pCommand>. Pointer on the
* command arguments is storead at the address pointed to by <pArgs>.
*
* RETURNS: OK if the command has bee correct parsed, ERROR otherwise.
*
*/

static int wrProxyClientCommandParse
    (
    char *	buffer,
    char **	pCommand,
    char **	pArgs
    )
    {
    int	commandLength = strlen (buffer);
    int	ix;

    /* Remove leading spaces */

    for (ix = 0; (ix < commandLength) && (isspace ((int)buffer[ix])); ix ++)
    	;

    /* A command is mandatory; return error if none was specified */

    if (ix == commandLength)
    	return (-1);

    *pCommand = &buffer[ix];

    /* Get end of command */

    for (; ix < commandLength; ix ++)
    	{
	if (isspace ((int) buffer[ix]))
	    {
	    buffer[ix] = EOS;
	    ix ++;

	    /* Remove any extra space character */

	    while (isspace((int) buffer[ix]))
	    	ix ++;
	    break;
	    }
	}

    /* Check if there are arguments to command */

    if (ix == commandLength)
    	*pArgs = NULL;
    else
    	*pArgs = &buffer[ix];

    return (0);
    }

/*******************************************************************************
*
* wrProxyClientAckSend - write command acknowledgement to the client
*
* This routine writes the specified buffer to the client. If <pProxy> is
* NULL, then the buffer is printed on stdout.
*
* RETURNS: OK or ERROR if there was an error writing data to the client.
*/

static STATUS wrProxyClientAckSend
    (
    PROXY_CONNECTION *  pProxy,		/* proxy connection */
    STATUS		result,		/* command result */
    char *		data		/* additional data to send to client */
    )
    {
    size_t	length;		/* data length */
    STATUS	status = OK;	/* command status */
    char	message[10];

    strcpy (message, result == OK ? "ok" : "error");

    if ((data != NULL) && (data[0] != EOS))
    	{
    	strcat (message, " ");
	length = strlen (message);	/* do not send EOS */
	}
    else
    	length = strlen (message) + 1;	/* do send EOS */


    /*
     * If the command is run from a startup script <pProxy == NULL>, we
     * print the result on stdout.
     */

    if (pProxy != NULL)
    	{
	if ((status = wrProxyClientWrite (pProxy, message, length)) == -1)
	    goto done;

	if ((data != NULL) && (data[0] != EOS))
	    {
	    length = strlen (data) + 1;
	    if ((status = wrProxyClientWrite (pProxy, data, length)) == -1)
		goto done;
	    }
	}
    else
    	{
	if ((data != NULL) && (data[0] != EOS))
	    fprintf (stdout, "%s%s\n", message, data);
	else
	    fprintf (stdout, "%s\n", message);
	}
done:
#ifdef	DEBUG
    if (status == ERROR)
	wrProxyErrnoPrint ("Error sending acknowledgement to client: ",
				NULL, 0);
#endif	/* DEBUG */
    return (status);
    }

/*******************************************************************************
*
* wrProxyClientWrite - write data to the client
*
* This routine writes the specified buffer to the client.
*
* RETURNS: number of bytes written to the client or ERROR
*/

static int wrProxyClientWrite
    (
    PROXY_CONNECTION *  pProxy,
    char *		buffer,
    int			nBytes
    )
    {
    int	bytesSent = 0;

    while (bytesSent != nBytes)
    	{
	int count;
	
	if ((count = send (pProxy->clientFd, buffer,
			nBytes - bytesSent, 0)) == -1)
	    {
	    if (errno == EINTR)
	    	continue;
	    else
		return (ERROR);
	    }
	bytesSent += count;
	buffer += count;
	}

    return (nBytes);
    }

/*******************************************************************************
*
* wrProxyClientRead - read data from the client
*
* This routine reads and handles datas from a client
*
* RETURNS: OK or ERROR if there was a fatal error reading from client.
*/

static STATUS wrProxyClientRead
    (
    PROXY_CONNECTION *  pProxy
    )
    {
    int		nBytes;
    int		packetSize;
    STATUS	status;

    if ((pProxy->status & PROXY_STATUS_CONNECTED) &&
    	(pProxy->mode == PROXY_MODE_PACKET))
    	{
	/* First get packet size */

	nBytes = recv (pProxy->clientFd, (char *)&packetSize, 4, 0);

	if ((nBytes == 0) || (nBytes == -1) || (nBytes != 4))
	    {
	    /*
	     * Connection with the client has been closed or
	     * is broken.
	     */

#ifdef	DEBUG
	    printf ("wrproxy: Error reading packet length from client " \
	    	    "(nBytes=%d, errno=%#x)\n", nBytes, errno);
#endif	/* DEBUG */

	    return (ERROR);
	    }

	packetSize = ntohl (packetSize);

	if (packetSize > PROXY_MAX_PACKET_SIZE)
	    {

	    /*
	     * This should not happen, there is a problem with the client;
	     * close connection.
	     */

#ifdef	DEBUG
	    printf ("wrproxy: packet size is too big (packetSize=%d)\n",
	    		packetSize);
#endif	/* DEBUG */
	    return (ERROR);
	    }

	nBytes = 0;
	while (nBytes != packetSize)
	    {
	    int count = recv (pProxy->clientFd, &buffer[nBytes],
						packetSize - nBytes, 0);

	    if ((count == -1) && (errno == EINTR))
	    	{
		/* In case of interrupted system call; ignore the error */

	    	continue;
		}
	    else if (count == 0 || count == -1)
	    	{
		/*
		 * There was an error reading data from the client; close
		 * connection.
		 */

#ifdef	DEBUG
		printf ("wrproxy: Error reading data from client " \
			"(count=%d, errno=%#x)\n", count, errno);
#endif	/* DEBUG */
		return (ERROR);
		}

	    nBytes += count;
	    }
	}
    else
    	{
	do
	    {
	    nBytes = recv (pProxy->clientFd,
	    			buffer, PROXY_MAX_PACKET_SIZE - 1, 0);
	    } while (nBytes == -1 && errno == EINTR);

	if (nBytes == -1 || nBytes == 0)
	    {
	    /*
	     * There was an error reading data from the client; 
	     * close connection.
	     */

#ifdef	DEBUG
	    printf ("wrproxy: Error reading data from client " \
	    	    "(nBytes=%d, errno=%#x)\n", nBytes, errno);
#endif	/* DEBUG */
	    return (ERROR);
	    }
	}

    /* If we are already connected to the target, simply forward the packet */

    if (pProxy->status & PROXY_STATUS_CONNECTED)
    	{
	if (wrProxyBackendWrite (pProxy, buffer, nBytes) == ERROR)
	    {
	    /*
	     * There was an error writting data to the client;
	     * close connection
	     */

#ifdef	DEBUG
	    printf ("wrproxy: Error sending data to client (errno=%#x)\n",
	    			errno);
#endif	/* DEBUG */
	    return (ERROR);
	    }

	return (OK);
	}

    /* Make sure command ends with a NULL character. */

    buffer[nBytes] = EOS;

    /* Run the command */

    status = wrProxyCommandRun (pProxy, buffer);

    return (status);
    }

/*******************************************************************************
*
* wrProxyCommandRun - execute a wrproxy command
*
* This routine executes the wrproxy command <command>. A pointer on the 
* result string will be stored in *<pResult>. If no information
* has to be returned by the command, *<pResult> is set to NULL.
*
* RETURNS: OK or ERROR if there was an error communicating with the client.
*/

static STATUS wrProxyCommandRun
    (
    PROXY_CONNECTION *	pProxy,	/* proxy that has run this command */
    char *		command	/* command to run */
    )
    {
    char *	cmdName;	/* command name */
    char *	args;		/* command arguments */
    size_t	nBytes;		/* command length */
    STATUS	status = ERROR;	/* command status */
    char *	result;		/* pointer on result string */

    /*
     * Clear command call back routine; it will be set by the run command
     * if necessary.
     */

    cmdHandleCallbackRtn = NULL;

    /* Compute again nBytes to get rid off ending NULL character */

    nBytes = strlen (command);

    for (;;)
    	{
	/* Remove any ending \n or \r character */

	if (nBytes &&
	    ((command [nBytes - 1] == '\n') || (command [nBytes - 1] == '\r')))
	    {
	    command[nBytes - 1] = EOS;
	    nBytes --;
	    }
	else
	    break;
	}

    /* Clear error message */

    errorMsg[0] = EOS;

    /* Clear *pResult */

    result = NULL;

    /* parse the command line */

    if (wrProxyClientCommandParse (command, &cmdName, &args) == -1)
    	{
	strncpy (errorMsg, CMDLINE_PARSE_ERROR, ERROR_MSG_LENGTH);
	result = errorMsg;
	goto done;
	}
    else
        {
	UINT32	ix;

	for (ix = 0; ix < NELEMENTS(proxyCmdArray); ix ++)
	    {
	    if (strcmp (cmdName, proxyCmdArray[ix].cmdName) == 0)
	    	{
		status = proxyCmdArray[ix].cmdHandleRtn (pProxy,
							 errorMsg,
							 ERROR_MSG_LENGTH,
							 &result,
							 args);
		goto done;
		}
	    }
	snprintf (errorMsg, ERROR_MSG_LENGTH, UNKNOWN_COMMAND, cmdName);
	result = errorMsg;
	}
done:
    /* Make sure errorMsg does end with a NULL character */

    errorMsg [ERROR_MSG_LENGTH - 1] = EOS;

    /* Send result to client */

    status = wrProxyClientAckSend (pProxy, status,
    					status == OK ? result : errorMsg);

    /* Call command call back routine if one is specified */

    if (cmdHandleCallbackRtn != NULL)
	cmdHandleCallbackRtn (result);

    return (status);
    }

/*******************************************************************************
*
* wrProxyModeArgGet - parse connect command arguments to retrieve mode value
*
* This routine parses the connect command arguments to retrieve the "mode"
* value; the value is stored at the address pointed to by <pMode>.
*
* RETURNS: OK if the mode was correctly parsed, ERROR otherwise
*/

static STATUS wrProxyModeArgGet
    (
    char *		args,		/* command arguments */
    char *		errorMsg,	/* where to store error message */
    int			errorMsgLength,	/* error message buffer length */
    PROXY_MODE *	pMode		/* where to store result */
    )
    {
    int		status;
    char	mode [30];

    status = wrProxyClientCommandArgGet (args, "mode=", NULL, mode, 30);
	
    if (status == -1)
    	{
	strncpy (errorMsg, MODE_PARSE_ERROR, errorMsgLength);
	return (ERROR);
	}
    else if (status == 1)
	{
	if (strcmp (mode, "raw") == 0)
	    *pMode = PROXY_MODE_RAW;
	else if (strcmp (mode, "packet") == 0)
	    *pMode = PROXY_MODE_PACKET;
	else
	    {
	    snprintf (errorMsg, errorMsgLength, UNKNOWN_MODE, mode);
	    return (ERROR);
	    }
	}
     else
     	{
     	/* Mode is not specified, default to raw mode */

	*pMode=PROXY_MODE_RAW;
	}
    return (OK);
    }

/*******************************************************************************
*
* wrProxyConnectCmdHandle - handle connect command
*
* This routine opens a new proxy connection. <args> is used to specify
* the target connection parameters, <pProxy> is used to specify the client
* that has sent the command; if <pProxy> is NULL, then it means this command
* is called from the startup script.
*
* RETURNS: OK or ERROR if the connect command has failed
*/

static STATUS wrProxyConnectCmdHandle
    (
    PROXY_CONNECTION *  pProxy,		/* proxy connection */
    char *		errorMsg,	/* where to store error message */
    int			errorMsgLength,	/* errorMsg buffer length */
    char **		pResult,	/* pointer on result string */
    char *		args		/* command arguments */
    )
    {
    if (pProxy == NULL)
    	{
	snprintf (errorMsg, errorMsgLength, STARTUP_SCRIPT_ERROR, "connect");
	return (ERROR);
	}

    pProxy->mode = PROXY_MODE_RAW;

    if (wrProxyModeArgGet (args, errorMsg, errorMsgLength,
    						&pProxy->mode) == ERROR)
	return (ERROR);
    else if (wrProxyBackendConnect (pProxy, args,
    				errorMsg, errorMsgLength, FALSE) == OK)
	{
	pProxy->status |= PROXY_STATUS_CONNECTED;
	return (OK);
	}
    else
	return (ERROR);
    }
    
#ifdef	HOST
/*******************************************************************************
*
* wrProxyClientCommandArgRemove - remove an argument from a command string
*
* This routine removes the argument <argName> from the command string <args>.
*
* RETURNS: N/A
*/

static void wrProxyClientCommandArgRemove
    (
    char *		args,		/* arguments to parse */
    const char *	argName		/* argument name */
    )
    {
    char *	start;
    char *	end;
    char *	ptr;
    size_t	length;

    /*
     * Verify if the argument list is not empty. If it is indicate we have
     * not found the specified argument.
     */

    if (args == NULL)
    	return;

    length = strlen (args) + 1;

    /* look for argument argName in argument list */
    
    ptr = args;
again:
    if ((start = strstr (ptr, argName)) == NULL)
    	return;
    else if ((start != (char *) args) &&
    	     (*(start - 1) != ';') &&
	     (!isspace ((int) *(start - 1))))
    	{
	/* Skip this occurence of the found string and look for another one */

	ptr = start + 1;
	goto again;
	}

    /* We have found the argument; remove it from string */

    if ((end = strchr (start, ';')) != NULL)
    	{
	/* Remove arg including ending ';' character */

	memmove (start, end + 1, (args + length) - (end + 1));
	}
    else
    	{
	/* Remmove arg including leading ';' characther if any */

	if (start != args)
	    *(start - 1) = EOS;
    	else
	    *start = EOS;
	}
    }

/*******************************************************************************
*
* wrProxyCmdCallback - call back routine for list command
*
* This routine is called to free memory allocated by the various command
* handler routines once it has been fully handled.
*
* RETURNS: N/A
*/

static void wrProxyCmdCallback
    (
    char *	result		/* list command result */
    )
    {
    if (result != NULL)
    	free (result);
    }

/*******************************************************************************
*
* wrProxyCreateCmdHandle - handle creation of new connection channel
*
* This routine handles the command to creates a new proxy connection channel.
* <args> is used to specify the target connection parameters,
* <pProxy> is used to specify the client
* that has sent the command; if <pProxy> is NULL, then it means this command
* is called from the startup script. If the create command is succesfull, a
* buffer will be allocated to store the result string and a pointer to this
* buffer will be stored in *<pResult>; this buffer has to be freed by the
* caller. If the command does fail, an error message is put in <errorMsg> 
* buffer.
*
* RETURNS: OK or ERROR if the create command has failed
*/

static STATUS wrProxyCreateCmdHandle
    (
    PROXY_CONNECTION *  pProxy,		/* proxy connection */
    char *		errorMsg,	/* where to store error message */
    int			errorMsgLength,	/* errorMsg buffer length */
    char **		pResult,	/* pointer on result string */
    char *		args		/* command arguments */
    )
    {
    PROXY_SERVER *	pServer = NULL;
    STATUS		status = OK;
    PROXY_MODE		mode;
    int			portNumber = 0;

    /* Validate parameters for future connect command. */

    if ((wrProxyModeArgGet (args, errorMsg,
    				ERROR_MSG_LENGTH, &mode) == ERROR) ||
        (wrProxyBackendConnect (pProxy, args,
    				errorMsg, errorMsgLength, TRUE) == ERROR))
	{
    	return (ERROR);
	}

    /* Check if a port number is specified by client and valid (a short) */

    status  = wrProxyClientCommandArgGet (args, PPORT_OPTION_NAME,
    						&portNumber, 0, 0);
	
    if (status == -1 || ((portNumber & 0xffff0000) != 0))
    	{
	strncpy (errorMsg, PPORT_PARSE_ERROR, errorMsgLength);
	return (ERROR);
	}
    else if (status == 1)
    	{
	/* Remove pport argument from the command string for better display */

	wrProxyClientCommandArgRemove (args, PPORT_OPTION_NAME);
	}

    /* Create the proxy server */

    if ((pServer = wrProxyServerCreate ((u_short) portNumber, 1, args,
				errorMsg, errorMsgLength)) == NULL)
    	return (ERROR);

    *pResult = malloc (NUMBER_STR_LENGTH + strlen (PPORT_OPTION_NAME) + 1);

    if (*pResult != NULL)
	sprintf (*pResult, "%s%d", PPORT_OPTION_NAME, pServer->portNumber);

    /* Add command call back to free memory stored to store result */

    cmdHandleCallbackRtn = wrProxyCmdCallback;

    return (OK);
    }

/*******************************************************************************
*
* wrProxyDeleteCmdHandle - handle deletion of a proxy connection channel
*
* This routine handles the command used to delete a proxy connection channel.
* <args> is used to specify the port number of the proxy server
* to delete, <pProxy> is used to
* specify the client that has sent the command; if <pProxy> is NULL,
* then it means this command is called from the startup script.
*
* RETURNS: OK or ERROR if the delete command has failed
*/

static STATUS wrProxyDeleteCmdHandle
    (
    PROXY_CONNECTION *  pProxy,		/* proxy connection */
    char *		errorMsg,	/* where to store error message */
    int			errorMsgLength,	/* errorMsg buffer length */
    char **		pResult,	/* pointer on result string */
    char *		args		/* command arguments */
    )
    {
    int			status;
    int			portNumber;
    PROXY_SERVER *	pServer;
    PROXY_CONNECTION *	pTmpProxy;

    status  = wrProxyClientCommandArgGet (args, PPORT_OPTION_NAME,
    						&portNumber, 0, 0);
	
    if (status == -1 || status == 0)
    	{
	strncpy (errorMsg, PPORT_PARSE_ERROR, errorMsgLength);
	return (ERROR);
	}

    for (pServer = pServerList;
	 pServer != NULL;
	 pServer = pServer->next)
	{
	if (pServer->portNumber == (u_short) portNumber)
	    break;
	}

    if (pServer == NULL)
    	{
	strncpy (errorMsg, CONNECTION_NOT_FOUND, errorMsgLength);
	return (ERROR);
	}

    if (pServerList == pServer)
    	pServerList = (PROXY_SERVER *) pServer->next;

    if (pServer->prev != NULL)
	((PROXY_SERVER *)pServer->prev)->next = pServer->next;

    if (pServer->cnctNum != 0)
    	{
	for (pTmpProxy = pProxyList; pTmpProxy != NULL;
		pTmpProxy = (PROXY_CONNECTION *) pTmpProxy->next)
	    {
	    if (pTmpProxy->pServer == pServer)
	    	{
		pTmpProxy->pServer = NULL;
		break;
		}
	    }
	}
    CLOSE_SOCKET (pServer->sockId);
    free (pServer->cnctArgs);
    free (pServer);

    return (OK);
    }

/*******************************************************************************
*
* wrProxyListCmdHandle - handle deletion of a proxy connection channel
*
* This routine handles the command used to delete a proxy connection channel.
* <args> is used to specify the port number of the proxy server
* to delete, <pProxy> is used to
* specify the client that has sent the command; if <pProxy> is NULL,
* then it means this command is called from the startup script.
*
* RETURNS: OK or ERROR if the delete command has failed
*/

static STATUS wrProxyListCmdHandle
    (
    PROXY_CONNECTION *  pProxy,		/* proxy connection */
    char *		errorMsg,	/* where to store error message */
    int			errorMsgLength,	/* errorMsg buffer length */
    char **		pResult,	/* pointer on result string */
    char *		args		/* command arguments */
    )
    {
    PROXY_SERVER *	pServer;
    size_t		length = 0;
    char *		ptr;

    for (pServer = pServerList;
	 pServer != NULL;
	 pServer = pServer->next)
	{
	if (pServer->cnctArgs != NULL)
	    {
	    /*
	     * Compute string length:
	     *	- connection arguments 
	     *  - ; character
	     *  - "pport=" string
	     *  - port number
	     *  - separator (| or EOS for last one).
	     */

	    length += strlen (pServer->cnctArgs) + 1 + 
	    		strlen (PPORT_OPTION_NAME) + NUMBER_STR_LENGTH + 1;
	    }
	}
    if ((*pResult = calloc (1, length)) == NULL)
    	{
	wrProxyErrnoPrint (MEM_ALLOC_FAILURE, errorMsg, errorMsgLength);
	return (ERROR);
	}

    ptr = *pResult;
    ptr[0] = EOS;

    /* Generate the string */

    for (pServer = pServerList;
	 pServer != NULL;
	 pServer = pServer->next)
	{
	if (pServer->cnctArgs == NULL)
	    continue;

	if (ptr != *pResult)
	    strcat (ptr++, "|");

	sprintf (ptr, "%s%d;%s", PPORT_OPTION_NAME,
			pServer->portNumber, pServer->cnctArgs);
	ptr += strlen (ptr);
	}

    /* Install list command call back to free memory when it is done */

    cmdHandleCallbackRtn = wrProxyCmdCallback;

    return (OK);
    }

/*******************************************************************************
*
* wrProxyInitStringRun - run the provided initialization string
*
* This routine runs the provided initialization string. Each command within
* this initialization string is separated by a new line character.
*
* RETURNS: OK or ERROR if there was an error reading from target or writing
*          to client.
*/

static STATUS wrProxyInitStringRun
    (
    char *	initString	/* initialization string to run */
    )
    {
    char *	start;
    char *	end;
    STATUS	status;

    if (initString == NULL)
    	return (OK);

    start = initString;

    fprintf (stdout, "Executing startup script...\n\n");

    while (start != NULL)
    	{
	if ((end = strchr (start, COMMAND_SEPARATOR)) != NULL)
	    *end = EOS;

	fprintf (stdout, "%s\n",start);

	status = wrProxyCommandRun (NULL, start);

	if (end == NULL)
	    break;
	else
	    start = end + 1;
	}
    return (OK);
    }
#endif	/* HOST */

/*******************************************************************************
*
* wrProxyTgtRead - read data from target and transmit it to client
*
* This routine reads datas from target and transmit it to client.
*
* RETURNS: OK or ERROR if there was an error reading from target or writing
*          to client.
*/

static STATUS wrProxyTgtRead
    (
    PROXY_CONNECTION *  pProxy	/* proxy connection to use */
    )
    {
    int			nBytes;

    nBytes = wrProxyBackendRead(pProxy, buffer, PROXY_MAX_PACKET_SIZE);

    /* If nBytes is 0, don't consider this as an error */

    if (nBytes == -1)
	return (ERROR);
    else if (nBytes == 0)
    	return (OK);

    if (pProxy->mode == PROXY_MODE_RAW) 
    	{
	if (wrProxyClientWrite (pProxy, buffer, nBytes) != nBytes)
	    return (ERROR);
	}
    else if (pProxy->mode == PROXY_MODE_PACKET)
    	{
	int	size;

	/*
	 * packet oriented connection first transmit the size of the
	 * packet.
	 */

	size = htonl (nBytes);
	
	if (wrProxyClientWrite (pProxy, (char *)&size, sizeof(int))
			!= sizeof(int))
	    return (ERROR);

	/* Now, we can transmit the packet */

	if (wrProxyClientWrite (pProxy, buffer, nBytes) != nBytes)
	    return (ERROR);
	}
    return (OK);
    }

/*******************************************************************************
*
* wrProxyServerCreate - create a new proxy server
*
* This routine opens a new proxy connection. <cnctMaxNum> parameter is used
* to specify the number of clients that can, at the same time, connect
* to this server; if the number is -1, this number is not limited. If
* <cnctArgs> argument is specified, it is used to specify the arguments of
* the wrproxy backend connection command that will be run automatically
* when a client connect to the created server, if it is set to NULL,
* then no backend connection command is called at server creation.
*
* RETURNS: a pointer on PROXY_SERVER structure that describes the newly
*	   created server or NULL if there was an error creating the server.
*/

static PROXY_SERVER * wrProxyServerCreate
    (
    u_short		portNumber,	/* port number to use */
    int			cnctMaxNum,	/* maximum number of connections */
    char *		cnctArgs,	/* target connection arguments */
    char *		errorMsg,	/* where to store error message */
    int			errorMsgLength	/* error message max length */
    )
    {
    int			sockId = -1;	/* socket used for clients */
					/* connections */
    struct sockaddr_in  addr;		/* Address information */
    PROXY_SERVER *	pServer = NULL;	/* proxy server */

    /* First make sure the port number is not already in use. */

    if (portNumber != 0)
    	{
	for (pServer = pServerList; pServer != NULL; pServer = pServer->next)
	    {
	    if (pServer->portNumber == portNumber)
	    	{
	    	strncpy (errorMsg, PPORT_ALREADY_IN_USE, errorMsgLength);
		return (NULL);
		}
	    }
	}
    pServer = NULL;	/* reset pServer pointer */

    /* Create and configure socket used for server */

    if ((sockId = socket (AF_INET, SOCK_STREAM, 0)) < 0)
    	{
	wrProxyErrnoPrint (SOCKET_CREATION_ERROR, errorMsg, errorMsgLength);
	return (NULL);
	}

    wrProxyTcpSocketOptionsSet (sockId);	/* set socket options */

    memset (&addr, 0, sizeof (addr));
    addr.sin_family		= AF_INET;
    addr.sin_addr.s_addr	= htonl(INADDR_ANY);
    addr.sin_port		= htons (portNumber);

    if (bind (sockId, (struct sockaddr *)&addr, sizeof(addr)) < 0)
    	{
	wrProxyErrnoPrint (SOCKET_BIND_ERROR, errorMsg, errorMsgLength);
	goto error;
	}

    /* Get port number used if it was not specified in command line */

    if (portNumber == 0)
    	{
	int opt = sizeof (addr);

	if (getsockname(sockId, (struct sockaddr *) &addr, &opt) != 0)
	    {
	    wrProxyErrnoPrint (GETSOCKNAME_ERROR, errorMsg, ERROR_MSG_LENGTH);
	    goto error;
	    }
	portNumber = ntohs (addr.sin_port);
	}

    /* Put socket in listen mode */

    if (listen(sockId, cnctMaxNum == -1 ? SOMAXCONN : 1) < 0)
	{
	wrProxyErrnoPrint (SOCKET_LISTEN_ERROR, errorMsg, errorMsgLength);
	goto error;
	}

    /* Create proxy server structure */

    if ((pServer = calloc (1, sizeof (PROXY_SERVER))) == NULL)
    	{
	wrProxyErrnoPrint (MEM_ALLOC_FAILURE, errorMsg, errorMsgLength);
	goto error;
	}

    pServer->sockId	= sockId;
    pServer->portNumber	= portNumber;
    pServer->cnctMaxNum	= cnctMaxNum;
    if (cnctArgs != NULL)
    	{
	if ((pServer->cnctArgs = strdup (cnctArgs)) == NULL)
	    {
	    wrProxyErrnoPrint (MEM_ALLOC_FAILURE, errorMsg, errorMsgLength);
	    goto error;
	    }
	}

    /* Add server to server list */

    if (pServerList == NULL)
	pServerList = pServer;
    else
    	{
	pServer->next	= (void *) pServerList;
	pServerList->prev	= (void *) pServer;
	pServerList		= (void *) pServer;
	}

    return (pServer);
error:
    if (pServer != NULL)
    	free (pServer->cnctArgs);
    free (pServer);
    if (sockId != -1)
    	CLOSE_SOCKET (sockId);
    return (NULL);
    }

/*******************************************************************************
*
* wrProxyConnectionOpen - open proxy connection
*
* This routine opens a new proxy connection.
*
* RETURNS: N/A
*/

static void wrProxyConnectionOpen
    (
    PROXY_SERVER *	pServer		/* server used for proxy connection */
    )
    {
    struct sockaddr_in  from;        /* Address information */
    int			fd;
    int			length;
    PROXY_CONNECTION *	pProxy;

    memset((char *)&from, 0, sizeof(from));
    length = sizeof (from);

    fd = accept (pServer->sockId, (struct sockaddr *) &from, &length);

    if (fd == -1)
    	{
	wrProxyErrnoPrint (SOCKET_ACCEPT_ERROR, NULL, 0);
	return;
	}

#ifdef	DEBUG
    printf ("wrproxy: Received a connection request from %s\n",
    					inet_ntoa (from.sin_addr));
#endif	/* DEBUG */

    wrProxyTcpSocketOptionsSet (pServer->sockId);	/* set socket options */

    if (pServer->cnctMaxNum != -1 && pServer->cnctNum == pServer->cnctMaxNum)
    	{
	/*
	 * For this server, we have already reached the maximum number of
	 * silmutaneous connections, decline this one.
	 */

	 CLOSE_SOCKET (fd);
	 return;
	 }

    /* setup proxy structure */

    pProxy = calloc (1, sizeof (PROXY_CONNECTION));

    if (pProxy == NULL)
	{
	wrProxyErrnoPrint (MEM_ALLOC_FAILURE, NULL, 0);
	CLOSE_SOCKET (fd);
	}
   else
	{
	pProxy->clientFd	= fd;
	pProxy->tgtFd		= -1;
	pProxy->pServer		= pServer;
	pServer->cnctNum	+= 1;

	if (pProxyList == NULL)
	    pProxyList = pProxy;
	else
	    {
	    pProxy->next = (void *) pProxyList;
	    pProxyList->prev = (void *) pProxy;
	    pProxyList = pProxy;
	    }

	if (pServer->cnctArgs != NULL)
	    {
	    /* Automatically connect to backend */

	    if ((wrProxyModeArgGet (pServer->cnctArgs,
				    errorMsg, ERROR_MSG_LENGTH,
				    &pProxy->mode) == ERROR) ||
		(wrProxyBackendConnect (pProxy,
			pServer->cnctArgs,
			errorMsg, ERROR_MSG_LENGTH, FALSE) == 0))
		{ 
		pProxy->status	|= PROXY_STATUS_CONNECTED;
		}
	    else
		{
		errorMsg [ERROR_MSG_LENGTH - 1] = EOS;
		wrProxyErrnoPrint (errorMsg, NULL, 0);
		wrProxyConnectionClose (pProxy);
		}
	    }
	}
    }

/*******************************************************************************
*
* wrProxyConnectionClose - close proxy connection
*
* This routine closes the specified proxy connection.
*
* RETURNS: N/A
*
* SEE ALSO:
*/

static void wrProxyConnectionClose
    (
    PROXY_CONNECTION *	pProxy	/* proxy connection to close */
    )
    {
#ifdef	DEBUG
    printf ("wrproxy: Close proxy connection %#x\n", (int) pProxy);
#endif	/* DEBUG */

    /* Remove node from list */

    if (pProxyList == pProxy)
    	pProxyList = (PROXY_CONNECTION *) pProxy->next;

    if (pProxy->prev != NULL)
	((PROXY_CONNECTION *)pProxy->prev)->next = pProxy->next;

    if (pProxy->next != NULL)
	((PROXY_CONNECTION *)pProxy->next)->prev = pProxy->prev;

    /* Close agent connection */

    if (pProxy->status & PROXY_STATUS_CONNECTED)
	wrProxyBackendClose (pProxy);

    /* Close client connection */

    CLOSE_SOCKET (pProxy->clientFd);

    /* Update server to indicate we are no longer connected */

    if (pProxy->pServer != NULL)
	pProxy->pServer->cnctNum --;

    /* Free buffer stored in pProxy structure */

    if (pProxy->buffer != NULL)
    	free (pProxy->buffer);

    /* free structure */

    free (pProxy);
    }

/*******************************************************************************
*
* wrProxyErrnoPrint - print errno
*
* This routine prints string associated with errno. If a buffer is provided
* the string is printed in this buffer otherwise it is printed on stderr.
*
* RETURNS: N/A
*
* NOMANUAL
*/

static void wrProxyErrnoPrint
    (
    char *	message,	/* additional message to print */
    char *	errorMsg,	/* where to store error message */
    int		errorMsgLength	/* error message max length */
    )
    {
#ifdef	WIN32
    LPVOID	lpMsgBuf;
    DWORD	dw = GetLastError(); 
    int		ix;

    FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER | 
        FORMAT_MESSAGE_FROM_SYSTEM,
        NULL,
        dw,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        (LPTSTR) &lpMsgBuf,
        0, NULL );

    ix = strlen (lpMsgBuf) - 1;
    while (ix > 0 && isspace ((int) ((char *)lpMsgBuf)[ix]))
	((char *)lpMsgBuf)[ix --] = EOS;
    if (errorMsg != NULL)
    	snprintf (errorMsg, errorMsgLength, "%s (%s)", message, lpMsgBuf);
    else
	fprintf (stderr, "%s (%s)\n", message, lpMsgBuf); 
    LocalFree(lpMsgBuf);
#else	/* WIN32 */
    if (errorMsg != NULL)
	snprintf (errorMsg, errorMsgLength, "%s (%s)",
					message, strerror(errno));
    else
	fprintf (stderr, "%s (%s)\n", message, strerror(errno));
#endif	/* WIN32 */
    }

/*******************************************************************************
*
* wrProxyTask - wrProxy server task
*
* This routine starts the proxy server
*
* RETURNS: N/A
*
* NOMANUAL
*/

void wrProxyTask (void)
    {
    fd_set		readFds;
    int			result;
    PROXY_CONNECTION *	pProxy;
    PROXY_CONNECTION *	pNextProxy;
    PROXY_SERVER *	pServer;

    for (;;)
    	{
	/*
	 * Reload the fd list; not that it would be possible to simply
	 * update this list when a new proxy connection would be created or
	 * deleted but it would make things more complex and less reliable.
	 */

	FD_ZERO (&readFds);

	for (pServer = pServerList; pServer != NULL; pServer = pServer->next)
	    {
	    FD_SET (pServer->sockId, &readFds);
	    }

	for (pProxy = pProxyList; pProxy != NULL;
		    pProxy = (PROXY_CONNECTION *) pProxy->next)
	    {
	    FD_SET (pProxy->clientFd, &readFds);
	    if (pProxy->tgtFd != -1)
		FD_SET ((u_int) pProxy->tgtFd, &readFds);
	    }

	/* Wait for requests or events */

	result = select (FD_SETSIZE, &readFds, (fd_set *) NULL,
			     (fd_set *) NULL, (struct timeval *) NULL);

	switch (result)
	    {
	    case 0:
	    case -1:
		continue;

	    default:
		{
		/* First check if this is a new client connection */

		for (pServer = pServerList;
		     pServer != NULL;
		     pServer = pServer->next)
		    {
		    if (FD_ISSET (pServer->sockId, &readFds))
			wrProxyConnectionOpen (pServer);
		    }

		/*
		 * Now check if we have received data from a connected
		 * agent or from a client
		 */

		for (pProxy = pProxyList; pProxy != NULL; pProxy = pNextProxy)
		    {
		    /*
		     * Store pointer on next proxy connection structure
		     * since the current connection may be closed during
		     * this loop.
		     */

		    pNextProxy = (PROXY_CONNECTION *) pProxy->next;

		    /* First check if data has been received from target */

		    if ((pProxy->tgtFd != -1) &&
		    		FD_ISSET((u_int) pProxy->tgtFd, &readFds))
			{
			if (wrProxyTgtRead (pProxy) == ERROR)
			    {
			    /*
			     * There was a fatal error; close client
			     * connection.
			     */
			    
			    wrProxyConnectionClose (pProxy);
			    continue;
			    }
			}

		    /* Now check if data has been received from a client */

		    if ((pProxy->clientFd != -1) &&
		    		FD_ISSET(pProxy->clientFd, &readFds))
			{
			/*
			 * wrProxyClientRead() will handle the packet received
			 * from the client.
			 */

			if (wrProxyClientRead (pProxy) == ERROR)
			    {
			    /*
			     * There was a fatal error; close client
			     * connection.
			     */

			    wrProxyConnectionClose (pProxy);
			    continue;
			    }
			}
		    }
		}
	    }
	}
    }

/*******************************************************************************
*
* wrProxyLibInit - Initialize proxy library
*
* This routine initializes the proxy library
*
* RETURNS: OK or ERROR on failure
*
*/

STATUS wrProxyLibInit
    (
#ifdef	HOST
    UINT32	portNumber,		/* proxy port number */
    char *	initString,		/* proxy initialization string */
    BOOL	hasControlPort		/* proxy has control port enabled? */
#else	/* HOST */
    UINT32	portNumber,		/* proxy port number */
    UINT32	priority,		/* proxy task priority */
    UINT32	stackSize,		/* proxy task stack size */
    UINT32	options			/* proxy task options */
#endif	/* HOST */
    )
    {
#ifndef	HOST
    int wrProxyTaskId;
#else	/* !HOST */
#ifdef	WIN32
    WSADATA		wsaData;	/* Windows Socket init data */

    if (WSAStartup (MAKEWORD (2, 2), &wsaData ) != 0)
	{
	wrProxyErrnoPrint ("Error starting WSA: ", NULL, 0);
	return (ERROR);
	}
#endif	/* WIN32 */
#endif	/* !HOST */

#ifdef	HOST
    if (hasControlPort)
    	{
#endif	/* HOST */
	if (wrProxyServerCreate ((u_short) portNumber, -1, 0, NULL, 0) == NULL)
	    return (ERROR);
#ifdef	HOST
	}
#endif	/* HOST */

#ifndef	HOST
    wrProxyTaskId = taskSpawn ("tWdbProxyTask", priority,
				   options, stackSize,
				   (FUNCPTR) wrProxyTask, 0, 0, 0,
				   0, 0, 0, 0, 0, 0, 0);

    return (wrProxyTaskId == ERROR ? ERROR : OK);
#else	/* HOST */
    /* Run startup script */

    if (initString != NULL)
	wrProxyInitStringRun (initString);

    wrProxyTask ();
    return (OK);
#endif	/* HOST */
    }
