/* wrProxyTcpSockLib.c - proxy TCP socket interface library */

/*
 * Copyright (c) 2005-2006 Wind River Systems, Inc.  
 *
 * The right to copy, distribute, modify or otherwise make use  of this 
 * software may be licensed only pursuant to the terms of an applicable Wind 
 * River license agreement.
 */

/*
modification history
--------------------
01b,16jun06,dbt  Added routine to validate backend arguments.
01a,20may05,dbt  written.
*/

/*
DESCRIPTION

This library contains the Target Agent Proxy interface for TCP socket 
library.
*/

#ifndef	HOST
#include "vxWorks.h"
#include "sockLib.h"
#include "hostLib.h"
#else	/* HOST */
#include <host.h>
#endif	/* !HOST */

#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <netdb.h>
#include <sys/types.h>
#include <string.h>
#ifndef	WIN32
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/tcp.h>
#else	/* WIN32 */
#include "rpc/rpc.h"
#include "io.h"
#endif	/* WIN32 */
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <unistd.h>


#ifdef	SUN4_SOLARIS2
#include <sys/sockio.h>
#include <sys/filio.h>
#endif /* SUN4_SOLARIS2 */

#include "wrProxyLibP.h"
#include "wrProxyBackendLib.h"

/* forward declarations */

static int	wrProxyTcpSockWrite	(PROXY_CONNECTION * pProxy,
					    char * buffer, int bufferLength);
static void	wrProxyTcpSockClose	(PROXY_CONNECTION * pProxy);
static int	wrProxyTcpSockRead	(PROXY_CONNECTION * pProxy,
					    char * buffer, int bufferLength);
static STATUS	wrProxyTcpSockOpen	(PROXY_CONNECTION * pProxy,
					    char * args, char *	errorMsg,
					    int errorMsgLength, 
					    BOOL dryRun);


/*******************************************************************************
*
* wrProxyTcpSockWrite - write a packet to TCP socket connection
*
* This routine writes a packet to TCP socket
*
* RETURNS: the number of bytes written or ERROR
*/


static int wrProxyTcpSockWrite
    (
    PROXY_CONNECTION *	pProxy,
    char *		buffer,
    int			bufferLength
    )
    {
    return (send (pProxy->tgtFd, buffer, bufferLength, 0));
    }

/*******************************************************************************
*
* wrProxyTcpSockClose - close a TCP socket connection
*
* This routine closes a TCP socket connection
*
* RETURNS: N/A
*/

static void wrProxyTcpSockClose
    (
    PROXY_CONNECTION *  pProxy
    )
    {
    CLOSE_SOCKET (pProxy->tgtFd);
    }

/*******************************************************************************
*
* wrProxyTcpSockRead - read data from a TCP socket connection
*
* This routine reads data from a TCP socket connection
*
* RETURNS: number of bytes read or ERROR
*/

static int wrProxyTcpSockRead
    (
    PROXY_CONNECTION *  pProxy,
    char *		buffer,
    int			bufferLength
    )
    {
    int	nBytes;

    nBytes = recv (pProxy->tgtFd, buffer, bufferLength, 0);

    /*
     * On a TCP connection, a value of 0 is returned when the peer has closed
     * the connection; we consider this as an error.
     */

    return (nBytes == 0 ? -1 : nBytes);
    }


/*******************************************************************************
*
* wrProxyTcpSockOpen - Open a new TCP socket connection
*
* This routine opens a new TCP socket connection
*
* RETURNS: OK if connection was succesfully opened, ERROR otherwise
*/

static STATUS wrProxyTcpSockOpen
    (
    PROXY_CONNECTION *	pProxy,
    char *		args,
    char *		errorMsg,
    int			errorMsgLength,
    BOOL		dryRun
    )
    {
    int			portNumber = 0;	/* target port number */
    struct hostent *	dest_host;	/* target address */
    u_long             	addr;           /* target address: long int format */
    char		tgtName [100];	/* target name */
    struct sockaddr_in 	srvAddr;
    int			status;
    int			dontBlock = 1;
    int			sockId;
    struct linger	optVal;
    int			on;

    status  = wrProxyClientCommandArgGet (args, "tgt=", NULL, tgtName, 100);
    
    if (status == ERROR)
    	{
	strncpy (errorMsg, "wrproxy: Unable to parse target name.",
				errorMsgLength);
	errorMsg[errorMsgLength - 1] = EOS;
	return (ERROR);
	}
    else if (status == 0)
    	{
	strncpy (errorMsg, "wrproxy: No target name is specified.",
				errorMsgLength);
	errorMsg[errorMsgLength - 1] = EOS;
	return (ERROR);
	}
  
    status  = wrProxyClientCommandArgGet (args, "port=", &portNumber, 0, 0);
    
    if (status == ERROR)
    	{
	strncpy (errorMsg, "wrproxy: Unable to parse port number.",
				errorMsgLength);
	errorMsg[errorMsgLength - 1] = EOS;
	return (ERROR);
	}
    else if (status == 0)
    	{
	strncpy (errorMsg, "wrproxy: No port number is specified.",
				errorMsgLength);
	errorMsg[errorMsgLength - 1] = EOS;
	return (ERROR);
	}

    /* Get server address */

    memset ((void *)&srvAddr, 0, sizeof(srvAddr));

    /* convert <tgtName> into the target board Internet address */

    if ((dest_host = (struct hostent *) gethostbyname (tgtName)) == NULL)
	{
	if ((addr = inet_addr (tgtName)) == (unsigned long) -1)
	    {
	    snprintf (errorMsg, errorMsgLength,
			    "wrproxy: Can not get IP address for %s.",
			    tgtName);
	    return (ERROR);
	    }
	srvAddr.sin_addr.s_addr = addr;
	}
    else
    	{
	srvAddr.sin_addr.s_addr = * ((u_long *) dest_host->h_addr_list[0]);
#ifndef	HOST
	hostentFree (dest_host);
#endif	/* HOST */
	}

    if (dryRun)
    	return (OK);

    srvAddr.sin_family	= AF_INET;
    srvAddr.sin_port	= htons ((u_short) portNumber);

    /* Create TCP socket */

    if ((sockId = socket (AF_INET, SOCK_STREAM, 0)) < 0)
	{
	strncpy (errorMsg, "wrproxy: Unable to create socket.",
			errorMsgLength);
	errorMsg[errorMsgLength - 1] = EOS;
	return (ERROR);
	}

    /* connect a socket */

    if (connect(sockId,(struct sockaddr *) &srvAddr,
    			sizeof (struct sockaddr_in)) < 0)
	{
	strncpy (errorMsg, "wrproxy: Unable to connect socket.",
			errorMsgLength);
	errorMsg[errorMsgLength - 1] = EOS;
	CLOSE_SOCKET (sockId);
        return(ERROR);
        }

    /* Set this socket to non-blocking */

#ifdef WIN32
    ioctlsocket (sockId, FIONBIO, &dontBlock);
#else /* WIN32 */
    ioctl (sockId, FIONBIO, (int) &dontBlock);
#endif /* WIN32 */

    /*
     * set SO_LINGER socket option on socket so that it closes the
     * connections gracefully, when required to close.
     */

    optVal.l_onoff = 1;
    optVal.l_linger = 0;

    setsockopt (sockId, SOL_SOCKET, SO_LINGER,
    				(void *)&optVal, sizeof (optVal));

    /* Set TCP_NODELAY socket option to optimize communication */

    on = 1;
    setsockopt(sockId, IPPROTO_TCP, TCP_NODELAY, (void *)&on, sizeof(on));

    /* Update target fd field in proxy structure */

    pProxy->tgtFd = sockId;

    return (0);
    }

/*******************************************************************************
*
* wrProxyTcpSockRegister - register TCP socket connection
*
* This routine registers TCP socket connection
*
* RETURNS: N/A
*/

void wrProxyTcpSockRegister (void)
    {
    wrProxyBackendRegister ("tcpsock", wrProxyTcpSockOpen,
			    wrProxyTcpSockRead, wrProxyTcpSockWrite,
			    wrProxyTcpSockClose);
    }
