/* wrProxyTipcSockLib.c - proxy TIPC socket interface library */

/*
 * Copyright (c) 2005-2006 Wind River Systems, Inc.  
 *
 * The right to copy, distribute, modify or otherwise make use  of this 
 * software may be licensed only pursuant to the terms of an applicable Wind 
 * River license agreement.
 */

/*
modification history
--------------------
01f,13jul06,dbt  Fixed error round by coverity.
01e,16jun06,dbt  Added routine to validate backend arguments.
01d,22aug05,dbt  Added support for Linux host.
01c,08jun05,dbt  Fixed build error.
01b,08jun05,dbt  Added support for specifying TIPC port type & instance.
01a,20may05,dbt  written.
*/

/*
DESCRIPTION

This library contains the Target Agent Proxy interface for TIPC socket 
library.
*/


/* TIPC backend is only supported on VxWorks or Linux */

#if	!defined(HOST) || (defined(HOST) && defined(LINUX))

/* includes */

#ifndef	HOST
#include <vxWorks.h>
#include <sockLib.h>
#include <hostLib.h>
#else	/* HOST */
#include <host.h>
#endif	/* HOST */
#include <tipc/tipc.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/ioctl.h>

#include "wrProxyLibP.h"
#include "wrProxyBackendLib.h"

/* forward declarations */

static int	wrProxyTipcSockWrite	(PROXY_CONNECTION * pProxy,
					    char * buffer, int bufferLength);
static void	wrProxyTipcSockClose	(PROXY_CONNECTION * pProxy);
static int	wrProxyTipcSockRead	(PROXY_CONNECTION * pProxy,
					    char * buffer, int bufferLength);
static STATUS	wrProxyTipcSockOpen	(PROXY_CONNECTION * pProxy,
					    char * args, char *	errorMsg,
					    int errorMsgLength, BOOL dryRun);

/*******************************************************************************
*
* wrProxyTipcSockWrite - write a packet to TIPC socket connection
*
* This routine writes a packet to TIPC socket
*
* RETURNS: the number of bytes written or ERROR
*/

static int wrProxyTipcSockWrite
    (
    PROXY_CONNECTION *	pProxy,
    char *		buffer,
    int			bufferLength
    )
    {
    int	nBytes;

    nBytes = sendto(pProxy->tgtFd, buffer, bufferLength, 0,
    			(struct sockaddr *)pProxy->backendArgs,
			sizeof(struct sockaddr_tipc));
    return (nBytes);
    }

/*******************************************************************************
*
* wrProxyTipcSockClose - close a TIPC socket connection
*
* This routine closes a TIPC socket connection
*
* RETURNS: N/A
*/

static void wrProxyTipcSockClose
    (
    PROXY_CONNECTION *  pProxy
    )
    {
    free (pProxy->backendArgs);
    CLOSE_SOCKET (pProxy->tgtFd);
    }

/*******************************************************************************
*
* wrProxyTipcSockRead - read data from a TIPC socket connection
*
* This routine reads data from a TIPC socket connection
*
* RETURNS: number of bytes read or ERROR
*/

static int wrProxyTipcSockRead
    (
    PROXY_CONNECTION *  pProxy,
    char *		buffer,
    int			bufferLength
    )
    {
    struct sockaddr_tipc	from;
    int				length = sizeof(struct sockaddr_tipc);
    int				nBytes;

    nBytes = recvfrom (pProxy->tgtFd, buffer, bufferLength, 0,
				 (struct sockaddr *)&from, &length);
    return (nBytes);
    }

/*******************************************************************************
*
* wrProxyTipcSockOpen - Open a new TIPC socket connection
*
* This routine opens a new TIPC socket connection
*
* RETURNS: OK if connection was succesfully opened, ERROR otherwise
*/

static STATUS wrProxyTipcSockOpen
    (
    PROXY_CONNECTION *	pProxy,
    char *		args,
    char *		errorMsg,
    int			errorMsgLength,
    BOOL		dryRun
    )
    {
    char			tgtName [100];	/* target address */
    struct sockaddr_tipc  *	pDstAddr;	/* server TIPC address */
    unsigned int		zone, cluster, node;	/* tipc information */
    char			extra;		/* misc value */
    int				dontBlock = 1;
    int				sockId;			/* TIPC socket */
    struct sockaddr_tipc	addr;		
    int				status;			/* misc status */
    UINT32			tipcPortType;		/* TIPC port type */
    UINT32			tipcPortInstance;	/* TIPC port instance */

    /* Parse arguments */

    status  = wrProxyClientCommandArgGet (args, "tgt=", NULL, tgtName, 100);

    if (status == ERROR)
    	{
	strncpy (errorMsg, "wrproxy: Unable to parse target name.",
				errorMsgLength);
	errorMsg[errorMsgLength - 1] = EOS;
	return (ERROR);
	}
    else if (status == 0)
    	{
	strncpy (errorMsg, "wrproxy: No target name is specified.",
				errorMsgLength);
	errorMsg[errorMsgLength - 1] = EOS;
	return (ERROR);
	}

    if (sscanf (tgtName, "%u.%u.%u%c", &zone, &cluster, &node, &extra) != 3)
    	{
	snprintf (errorMsg, errorMsgLength,
			"wrproxy: Invalid TIPC address format (%s).\n",
			tgtName);
        return (ERROR);
	}

    status  = wrProxyClientCommandArgGet (args, "tipcpt=",
    				(int *) &tipcPortType, NULL, 0);
    if (status == ERROR)
    	{
	strncpy (errorMsg, "wrproxy: Unable to parse TIPC port type.",
				errorMsgLength);
	errorMsg[errorMsgLength - 1] = EOS;
	return (ERROR);
	}
    else if (status == 0)
    	{
	strncpy (errorMsg, "wrproxy: No TIPC port type specified.",
				errorMsgLength);
	errorMsg[errorMsgLength - 1] = EOS;
	return (ERROR);
	}
 
    status  = wrProxyClientCommandArgGet (args, "tipcpi=",
    				(int *) &tipcPortInstance, NULL, 0);
    if (status == ERROR)
    	{
	strncpy (errorMsg, "wrproxy: Unable to parse TIPC port instance.",
				errorMsgLength);
	errorMsg[errorMsgLength - 1] = EOS;
	return (ERROR);
	}
    else if (status == 0)
    	{
	strncpy (errorMsg, "wrproxy: No TIPC port instance specified.",
				errorMsgLength);
	errorMsg[errorMsgLength - 1] = EOS;
	return (ERROR);
	}

    if (dryRun)
    	return (OK);

    /* Save target TIPC address */
        
    pProxy->backendArgs = calloc (1, sizeof (struct sockaddr_tipc));
    if (pProxy->backendArgs == NULL)
    	{
	strncpy (errorMsg, "wrproxy: Unable to allocate memory.",
				errorMsgLength);
	errorMsg[errorMsgLength - 1] = EOS;
	return (ERROR);
	}
    pDstAddr = (struct sockaddr_tipc  *) pProxy->backendArgs;

    pDstAddr->family = AF_TIPC;
    pDstAddr->addrtype = TIPC_ADDR_NAME;
    pDstAddr->scope = TIPC_NODE_SCOPE;
    pDstAddr->addr.name.name.type = tipcPortType;
    pDstAddr->addr.name.name.instance = tipcPortInstance;
    pDstAddr->addr.name.domain = tipc_addr (zone, cluster, node);

    /* Open a datagram TIPC (unreliable, not connected) socket */

    if ((sockId = socket (AF_TIPC, SOCK_DGRAM, 0)) < 0)
	{
	strncpy (errorMsg, "wrproxy: Unable to create TIPC socket.",
			errorMsgLength);
	errorMsg[errorMsgLength - 1] = EOS;
	free (pProxy->backendArgs);
	return (ERROR);
	}

    /* Set this socket to non-blocking */

    ioctl (sockId, FIONBIO, (int) &dontBlock);

    /* Bind socket to specific port */

    bzero ((char *)&addr, sizeof(addr));

    addr.family = AF_TIPC;
    addr.addrtype = TIPC_ADDR_NAME;
    addr.scope = TIPC_NODE_SCOPE;
    addr.addr.name.name.type = tipcPortType;
    addr.addr.name.name.instance = tipcPortInstance;
    addr.addr.name.domain = tipc_addr (zone, cluster, node);

    if (bind (sockId, (struct sockaddr *)&addr, sizeof(addr)) != OK)
	{
	strncpy (errorMsg, "wrproxy: Unable to bind to TIPC WDB port.",
			errorMsgLength);
	errorMsg[errorMsgLength - 1] = EOS;
	free (pProxy->backendArgs);
	CLOSE_SOCKET (sockId);
	return (ERROR);
	}

    pProxy->tgtFd = sockId;

    return (OK);
    }

/*******************************************************************************
*
* wrProxyTipcSockRegister - register TIPC socket connection
*
* This routine registers TIPC socket connection
*
* RETURNS: N/A
*/

void wrProxyTipcSockRegister (void)
    {
    wrProxyBackendRegister ("tipcsock", wrProxyTipcSockOpen,
			    wrProxyTipcSockRead, wrProxyTipcSockWrite,
			    wrProxyTipcSockClose);
    }
#endif /* !defined(HOST) || (defined(HOST) && defined(LINUX)) */
