/* wrProxyUdpSockLib.c - proxy UDP socket interface library */

/*
 * Copyright (c) 2005-2006 Wind River Systems, Inc.  
 *
 * The right to copy, distribute, modify or otherwise make use  of this 
 * software may be licensed only pursuant to the terms of an applicable Wind 
 * River license agreement.
 */

/*
modification history
--------------------
01b,16jun06,dbt  Added routine to validate backend arguments.
		 Handle ECONNREFUSED errno in receive routine.
01a,20may05,dbt  written.
*/

/*
DESCRIPTION

This library contains the Target Agent Proxy interface for UDP socket 
library.
*/

#ifndef	HOST
#include <vxWorks.h>
#include <sockLib.h>
#include <hostLib.h>
#else	/* !HOST */
#include <host.h>
#endif	/* HOST */

#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <sys/types.h>
#ifndef	WIN32
#   include <sys/socket.h>
#   include <sys/ioctl.h>
#else	/* WIN32 */
#   include <rpc/rpc.h>
#   include <io.h>
#endif	/* WIN32 */
#include <arpa/inet.h>
#include <fcntl.h>
#include <unistd.h>

#ifdef	SUN4_SOLARIS2
#include <sys/sockio.h>
#include <sys/filio.h>
#endif /* SUN4_SOLARIS2 */

#include "wrProxyLibP.h"
#include "wrProxyBackendLib.h"

/* forward declarations */

static int	wrProxyUdpSockWrite	(PROXY_CONNECTION * pProxy,
					    char * buffer, int bufferLength);
static void	wrProxyUdpSockClose	(PROXY_CONNECTION * pProxy);
static int	wrProxyUdpSockRead	(PROXY_CONNECTION * pProxy,
					    char * buffer, int bufferLength);
static STATUS	wrProxyUdpSockOpen	(PROXY_CONNECTION * pProxy,
					    char * args, char *	errorMsg,
					    int errorMsgLength, BOOL dryRun);

/*******************************************************************************
*
* wrProxyUdpSockWrite - write a packet to UDP socket connection
*
* This routine writes a packet to UDP socket
*
* RETURNS: the number of bytes written or ERROR
*/

static int wrProxyUdpSockWrite
    (
    PROXY_CONNECTION *	pProxy,
    char *		buffer,
    int			bufferLength
    )
    {
    return (send (pProxy->tgtFd, buffer, bufferLength, 0));
    }

/*******************************************************************************
*
* wrProxyUdpSockClose - close a UDP socket connection
*
* This routine closes a UDP socket connection
*
* RETURNS: N/A
*/

static void wrProxyUdpSockClose
    (
    PROXY_CONNECTION *  pProxy
    )
    {
    CLOSE_SOCKET (pProxy->tgtFd);
    }

/*******************************************************************************
*
* wrProxyUdpSockRead - read data from a UDP socket connection
*
* This routine reads data from a UDP socket connection
*
* RETURNS: number of bytes read or ERROR
*/

static int wrProxyUdpSockRead
    (
    PROXY_CONNECTION *  pProxy,
    char *		buffer,
    int			bufferLength
    )
    {
    int	nBytes;

    nBytes = recv (pProxy->tgtFd, buffer, bufferLength, 0);

    /*
     * For a UDP connected socket, we may receive a ECONNREFUSED error if the
     * ARP entry of the target has been removed (if it has been rebooted e.g.),
     * we do not want this to impact the client (this is usually not 
     * expected with a UDP socket.)
     */

#ifdef	ECONNREFUSED
    if (nBytes == -1 && errno == ECONNREFUSED)
#elif	defined(WSAECONNREFUSED)
    if (nBytes == -1 && errno == WSAECONNREFUSED)
#endif
    	return (0);

    return (nBytes);
    }

/*******************************************************************************
*
* wrProxyUdpSockOpen - Open a new UDP socket connection
*
* This routine opens a new UDP socket connection
*
* RETURNS: OK if connection was succesfully opened, ERROR otherwise
*/

static STATUS wrProxyUdpSockOpen
    (
    PROXY_CONNECTION *	pProxy,
    char *		args,
    char *		errorMsg,
    int			errorMsgLength,
    BOOL		dryRun
    )
    {
    int			portNumber = 0;	/* target port number */
    struct hostent *	dest_host;	/* target address */
    u_long             	addr;           /* target address: long int format */
    char		tgtName [100];	/* target name */
    struct sockaddr_in 	srvAddr;
    int			status;
    int			dontBlock = 1;
    int			sockId;

    status  = wrProxyClientCommandArgGet (args, "tgt=", NULL, tgtName, 100);
    
    if (status == ERROR)
    	{
	strncpy (errorMsg, "wrproxy: Unable to parse target name.",
				errorMsgLength);
	errorMsg[errorMsgLength - 1] = EOS;
	return (ERROR);
	}
    else if (status == 0)
    	{
	strncpy (errorMsg, "wrproxy: No target name is specified.",
				errorMsgLength);
	errorMsg[errorMsgLength - 1] = EOS;
	return (ERROR);
	}
  
    status  = wrProxyClientCommandArgGet (args, "port=", &portNumber, 0, 0);
    
    if (status == ERROR)
    	{
	strncpy (errorMsg, "wrproxy: Unable to parse port number.",
				errorMsgLength);
	errorMsg[errorMsgLength - 1] = EOS;
	return (ERROR);
	}
    else if (status == 0)
    	{
	strncpy (errorMsg, "wrproxy: No port number is specified.",
				errorMsgLength);
	errorMsg[errorMsgLength - 1] = EOS;
	return (ERROR);
	}

    /* Get server address */

    memset ((void *)&srvAddr, 0, sizeof(srvAddr));

    /* convert <tgtName> into the target board Internet address */

    if ((dest_host = (struct hostent *) gethostbyname (tgtName)) == NULL)
	{
	if ((addr = inet_addr (tgtName)) == (unsigned long) -1)
	    {
	    snprintf (errorMsg, errorMsgLength,
			    "wrproxy: Can not get IP address for %s.",
			    tgtName);
	    return (ERROR);
	    }
	srvAddr.sin_addr.s_addr = addr;
	}
    else
    	{
	srvAddr.sin_addr.s_addr = * ((u_long *) dest_host->h_addr_list[0]);
#ifndef	HOST
	hostentFree (dest_host);
#endif	/* HOST */
	}

    if (dryRun)
    	return (OK);

    srvAddr.sin_family	= AF_INET;
    srvAddr.sin_port	= htons ((u_short) portNumber);

    /* Create the socket */

    if ((sockId = socket (AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
	{
	strncpy (errorMsg, "wrproxy: Unable to create socket.",
			errorMsgLength);
	errorMsg[errorMsgLength - 1] = EOS;
	return (ERROR);
	}

    /* Set this socket to non-blocking */

#ifdef WIN32
    ioctlsocket (sockId, FIONBIO, &dontBlock);
#else /* WIN32 */
    ioctl (sockId, FIONBIO, (int) &dontBlock);
#endif /* WIN32 */

    /* connect to server to filter UDP packets */

    if (connect (sockId, (struct sockaddr *)&srvAddr, sizeof(srvAddr)) != OK)
	{
	strncpy (errorMsg, "wrproxy: Unable to connect socket.",
					errorMsgLength);
	errorMsg[errorMsgLength - 1] = EOS;
	CLOSE_SOCKET (sockId);
	return (ERROR);
	}

    pProxy->tgtFd = sockId;

    return (OK);
    }

/*******************************************************************************
*
* wrProxyUdpSockRegister - register UDP socket connection
*
* This routine registers UDP socket connection
*
* RETURNS: N/A
*/

void wrProxyUdpSockRegister (void)
    {
    wrProxyBackendRegister ("udpsock", wrProxyUdpSockOpen,
			    wrProxyUdpSockRead, wrProxyUdpSockWrite,
			    wrProxyUdpSockClose);
    }
