/* wrproxy.c - the Wind River multi-protocol proxy */

/*
 * Copyright (c) 2005-2006 Wind River Systems, Inc.
 *
 * The right to copy, distribute or otherwise make use of this software may be
 * licensed only pursuant to the terms of an applicable Wind River license
 * agreement. No license to Wind River intellectual property rights is granted
 * herein. All rights not licensed by Wind River are reserved by Wind River.
 */

/*
modification history
--------------------
01f,05oct06,vhe  Set tipc code conditional.
01e,10aug06,dbt  Ignore SIGPIPE signal (Defect CQ:WIND00061631).
01d,21jun06,dbt  Added support for startup file.
		 Added support for -nocontrol option
01c,23sep05,xav  added port value check and removed warning.
01b,21sep05,xav  added dynamic UDP port and usage.
01a,24aug05,dbt  written
*/

/*
SYNOPSIS
\cs
wrproxy  [-p.ort <proxy tcp control port number>] [-s.tartup <startup file>] [-nocontrol] [-h.elp]
\ce

DESCRIPTION

The Wind River Proxy is a proxy that can be used as a gateway between Wind
River Workbench tools and the debugged system. The client (Wind River
Workbench tools) connect to the proxy using a network TCP/IP socket and
the proxy does connect to the debugged system using a configurable link
(TCP/IP socket, UDP/IP socket, TIPC).

Ex: running proxy on localhost

\cs
   Host
____________
|  Host    |
|  Tool    |
|__ ||_____|
|   ||     |
|   \/     |                 Target
|   TCP    |              ____________
|   ||     |              |           |
|__ ||_____|              |  Target   |
|   \/     |              |           |
| WR Proxy | ==> TIPC ==> |Debug agent|
|__________|              |___________|

\ce

In this example, the Host Tool uses TCP to talk to Wind River Proxy
which uses TIPC network protocol to communicate with the debug agent on the
target.

Ex: running proxy on target

\cs
   Host                      Target
____________              _____________
|          |              |           |
|  Host    | ==> TCP ==>  | WR Proxy  |
|  Tool    |              |___||______|
|__________|              |   ||      |
                          |   \/      |
                          |  UDP      |
                          |   ||      |
                          |   \/      |
                          |Debug agent|
                          |___________|
\ce

In this example, the target connection uses TCP to talk to Wind River Proxy
which uses UDP to communicate with the debug agent on the target.

GOALS

Using the Wind River Proxy is not mandatory for debug scenari but it is
useful in several cases:

\ms
\m 1. When the debugged system is not accessible from the host used to
run Wind River Workbench tools because it is on a private network; it is
possible to run the Wind River Proxy on a machine that is accessible from
Wind River Workbench tools network and has access to the debugged system
network and use this proxy to connect Wind River Workbench tools to the
debugged system.

\m 2. When the debugged system is behind a firewall, the Wind River Proxy
can be used to go through the firewall.

\m 3. When the debugged system is using TIPC to communicate. In this case,
the Wind River Proxy can be used as a gateway between the IP network
and the TIPC network (it has to run on a machine that has access to the
both networks). Note that this facility is available only on a Linux or
VxWorks host.

\m 4. The Wind River Proxy does allow some specific configurations like
connecting the Wind River Workbench tools to a target serial line using
a port server.

\me

In order to connect Wind River Workbench tools to a target using the
Wind River Proxy, a connection must be created in the proxy and the Host
tools can then use this connection to communicate to the target. Once
the connection is established, the Wind River Proxy does forward all
the packets received from the Host Tools to the target and vice-versa.
Note that some specific connnection configurations does allow the proxy
to encode/decode packets send to/receive from the target transparently
for the Host Tool; this is described later in this documentation. The
Wind River Proxy connections can be created either through the Wind River
Proxy control port or using the Wind River Proxy startup script. For
security reasons, it is possible to disable the Wind River Proxy control
port using the <-nocontrol> option; when this is done the startup script
is the only method to create new Wind River Proxy connections.


PROXY COMMANDS
There are several commands that can be used with the Wind River proxy;
those commands are string of the form:
\cs
<command> <param1>=xxx;<param2>=xxx;<param3>=xxx
\ce
The result of the commands is a string of the form:
\cs
ok <value1>=xxx;<value2>=xxx
\ce
or
\cs
error <error string>
\ce

The following commands are available:
\is
\i `list'
List all available Wind River Proxy connections.
\i `create'
Create a new Wind River Proxy connection.
\i `delete'
Delete an existing Wind River Proxy connection.
\i `connect'
Create a new Wind River Proxy connection and automatically connect to it.
\ie

The `create' command is used to create a new proxy connection to a target
and will assign a Wind River Proxy TCP port number to be able to access
to this connection. The connection with the target will be established
only when the client will connect to the specified TCP port number.

The `connect' command is also used to create a new proxy connection to a
target but the connection with the target will be established immediatly;
this command can not be called from the startup script. Once the `connect'
command has been sucessfully run, all packets sent to this connection
will be directory re-routed to the target and vice-versa.

The `list' command is used to list the already created connections with
their parameters.

The `delete' command is used to delete a previously created connection.

For both the `create' and `connect' commands, you must specify various
information about how you want to connect to the target. You must first
specify the target connection type you want to use using the "type="
string. You must specify "type=udpsock" for a UDP socket connection;
"type=tcpsock" for a TCP socket connection (this is not supported for
a Wind River Proxy running on VxWorks) and "type=tipcsock" for a TIPC
socket connection (note that this is available only on Linux hosts).

If the target connection type is udpsock or tcpsock, you must specify the
target IP address or target name using "tgt=<target name or ip address>",
the target UDP|TCP port number using "port=<port number>". For example, to
create a connection to target <t80-205>, using UDP port number 0x1234,
you can use the command:
\cs
create type=udpsock;tgt=t80-205;port=0x1234
\ce
or
\cs
connect type=udpsock;tgt=t80-205;port=0x1234
\ce

If the target connection type is tipcsock, you must specify the target
TIPC address using the "tgt=<target tipc address>" parameter, the TIPC
port type using the "tipcpt=<tipc port type>" parameter and the TIPC port
instance using the "tipcpi=<tipc port instance>" parameter. The command
should look like this:
\cs
create type=tipcsock;tgt=1.1.6;tipcpt=80;tipcpi=81
\ce
or
\cs
connect type=tipcsock;tgt=1.1.6;tipcpt=80;tipcpi=81
\ce

For the `create' command, you can also specify which port you want to be
used by the proxy to establish the connection to the target using the
"pport=<proxy port number>" parameter. If none is specified, the proxy
will pickup automatically an available port. The port used for the target
connection will be returned as a result of the proxy command:
\cs
create type=udpsock;tgt=t80-205;port=0x1234
ok pport=47346
create type=udpsock;tgt=t80-205;port=0x1234;pport=2000
ok pport=2000
\ce

For both the `create' and `connect' commands, you can also use the
`mode=raw|packet' and `protocol=raw|wdbserial' to describe the target
connection; those parameters are not dependent on the target connection
type you are using. The `mode=packet' option is used to indicate that
all the packets sent between the client and the Wind River Proxy when the
connection is established with the target will be preceded by the packet
size; you client must be adapted to first read/send packet size before
reading/sending a packet. If the `mode=raw' option is specified or no
`mode=xxx' option is specified, then the packet size is not transmitted
between the client and the proxy. The `protocol=wdbserial' parameter is
used to connect to a VxWorks target using the WDB serial backend when you
are not using the wdbserial backend on the host. This can be useful when
you want to connect the Wind River Target Server to a WDB agent running on
the target and using the serial line and if this serial line is connected
to a port server you can connect using a TCP port. When this parameter
is specified, some special characters expected by the WDB serial backend
are added to the packets sent to the port server TCP port and removed
from the packets received from the port server TCP port.

Note that when you use the `create' command, the resources to access the
target are used only when the client creates opens the proxy TCP port
established for this connection and the resources are released when
the connection is closed. It means that you can create two different
connections using the same target resources without any problem but you
may face issues if you try to connect to both of them at the same time.

For the `delete' command, you must specify the Wind River Proxy port number used by the connection you want to delete using the `pport=<port number>' option. Note that only a connection created using the `create' command can be deleted using the `delete' command.

The `list' command does not accept any parameter and simply list the proxy connections created. Each connection will be separaterd with a `|' character:
\cs
list
ok pport=2000;type=udpsock;tgt=t80-205;port=0x1234|pport=47346;type=udpsock;tgt=t80-205;port=0x1234
\ce

For all commands, unknown parameters are ignored; this is not considered
as an error. In addition, the client should not make any assumption on
the number of values returned by the command this could be changed in
the future. For example, the `create' command will always return the
value for <pport> but additional information may be returned in a future
version of the Wind River proxy.


STARTING PROXY

This process has to be launched manually on any host that has access to the 
debugged system.

COMMAND LINE EXECUTABLE

The proxy command line executable is named wrproxy on Unix system and
wrproxy.exe on Windows one.

OPTIONS
\is
\i `-p <port number> | -port <port number>'
Specify the TCP listening port number (default value is 0x4444)
\i `-h | -help'
Display a help message summarizing proxy usage and options
\i `-v | -version'
Display the proxy server version
\i `-nocontrol'
Disable the proxy control port, all connections have to be specified in
startup script.
\i `-s <startup script> | -startup <startup script>'
Run the specified startup script when the proxy is started.
\ie

EXAMPLES
Start Wind River Proxy on default port
\cs
    wrproxy
\ce
Start Wind River Proxy on TCP port number 1122
\cs
    wrproxy -p 1122
\ce

SUPPORT:

This proxy only supports TCP, UDP and TIPC protocols.

INCLUDES: N/A

SEE ALSO
`tgtsvr',
`usermode-ptrace' (UNIX only),
<WorkBench User's Guide>

NOROUTINES


*/

/* includes */

#include <host.h>
#include <stdio.h>
#ifdef	WIN32
#include <windows.h>
#endif	/* WIN32 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>
#include <string.h>
#include <signal.h>

#include "wrProxyLib.h"

/* defines */

#define COMMENT_MARKER '#'	/* single line comment marker */
#define COMMAND_SEPARATOR '\n'	/* single command separator */

/* local variables */

static BOOL	verboseMode = FALSE;
static UINT16	listeningTcpPort = 0x4444;
static char *	startupFileName = NULL;
static BOOL	hasControlPort = TRUE;	

/******************************************************************************
*
* usage - Print proxy usage
*
* This routine prints the proxy usage.
*
* RETURNS : N/A
*
* NOMANUAL
*/

static void usage
    (
    char * exeName
    )
    {
    fprintf(stderr, "Wind River proxy usage: %s [options] ...\n\n",exeName);
    fprintf(stderr, "Options:\n");
    fprintf(stderr, "   -p.ort <port number>\tChooses an alternate TCP \
control port (default is 0x%x).\n",0x4444);
    fprintf(stderr, "   -V\t\t\tActivates verbose mode.\n");    
    fprintf(stderr, "   -v.ersion\t\tPrint version and exit.\n");
    fprintf(stderr, "   -h.elp\t\tDisplays this message.\n");
    fprintf(stderr, "   -nocontrol\t\tDisable control port.\n");
    fprintf(stderr, "   -s.tartup <startup file>\tExecute commands in \
<startup file> at startup.\n");
    }

/******************************************************************************
*
* cmdLineParse - parse the command line.
*
* This routine parses the command line.
*
* RETURNS : OK or ERROR if the command line is wrong.
*
* NOMANUAL
*/

static STATUS cmdLineParse
    (
    int	   argc, 
    char * argv[]
    )
    {
    INT32 argNum = 1;
    BOOL  success;

    while (argNum < argc)
	{
	success = FALSE;

	if (strcmp(argv[argNum], "-v") == 0 ||
	    strcmp(argv[argNum], "-version") == 0)
	    {
	    fprintf (stderr, "Wind River proxy version %s.\n", WRPROXY_VERSION);
	    exit (0);
	    }

	/* -port option */
	
	if (strcmp(argv[argNum], "-port") == 0 ||
	    strcmp(argv[argNum], "-p") == 0)
	    {
	    int		 scan;
	    unsigned int port; /* scanf doesn't read shorts */
	    
	    if (++argNum == argc)
		goto error;
	    
	    if ((scan = sscanf(argv[argNum],"0x%x", &port)) != 1)
		{
		if ((scan = sscanf(argv[argNum],"%u", &port)) != 1)
		    {
		    fprintf (stderr, 
		    		"Can not understand port parameter %s\n",
				argv[argNum]);
		    goto error;
		    }
		}

	    if (port < 1 || port > 0x8000)
		{
		fprintf (stderr,
				"Can not use port 0x%x since\
					it's not in range (1-0x8000)\n", port);
		goto error;
		}

	    listeningTcpPort = port;
	    success          = TRUE;
	    }

	if (strcmp(argv[argNum], "-V") == 0)
	    {
	    verboseMode = TRUE;		/* activates verbose mode */
	    success = TRUE;
	    }
	
	if (strcmp(argv[argNum], "-nocontrol") == 0)
	    {
	    hasControlPort = FALSE;	/* disable control port */
	    success = TRUE;
	    }
	
	if (strcmp(argv[argNum], "-help") == 0 ||
	    strcmp(argv[argNum], "-h") == 0)
	    goto error;
		
	if (strcmp(argv[argNum], "--help") == 0 ||
	    strcmp(argv[argNum], "--h") == 0)
	    goto error;
		
	/* -s option */
	
	if (strcmp(argv[argNum], "-s") == 0 ||
	    strcmp(argv[argNum], "-startup") == 0)
	    {
	    if (++argNum == argc)
		goto error;
	   
	    startupFileName = argv[argNum];
	    success          = TRUE;
	    }
	if (!success)
	    {
	    fprintf (stderr, "Invalid parameters \"%s\"\n", argv[argNum]);	
	    goto error;
	    }
		
	argNum++;
	}

    return OK;

error:
    fprintf (stderr, "\n");
    usage (argv[0]);
    return ERROR;
    }

/******************************************************************************
*
* wrproxyStartupFileParse - parses the startup script
*
* This routine parses the startup script and generate a string that holds
* all the commands found in the startup script; each command is separated
* by a new line character. A pointer on the created string is stored in
* *<pInitString>. This memory has to be released by the caller.
*
* RETURNS : OK or ERROR if there was an error parsing the startup script.
*
* NOMANUAL
*/

static STATUS wrproxyStartupFileParse
    (
    char *	fileName,	/* startup file to parse */
    char **	pInitString	/* where to store pointer on init string */
    )
    {
    int		fd;		/* file descriptor for startup file */
    struct stat	statbuf;	/* to get info on the file */
    char *	ptr;
    char *	src;
    char *	dst;
    char *	fileContents = NULL;
    int		bytes2Read;
    int		inComment = 0;
    int		nBytes;

    /* get the file size */

    if (stat(fileName, &statbuf) < 0 ||
        ((fd = open (fileName, O_RDONLY)) < 0))
	{
	fprintf (stderr, "Error: unable to access file \"%s\"\n", fileName);
	return (-1);
	}

    /* Allocate buffers */

    if ((fileContents = (char *)malloc(statbuf.st_size+1)) == NULL )
	{
	fprintf (stderr,
		    "Error: unable to allocate memory to read file \"%s\"\n",
		    fileName);
	return (-1);
	}


    /* Read the full size and store it in fileContents buffer */

    bytes2Read	= statbuf.st_size;
    ptr		= fileContents;
    while ((nBytes = read(fd, ptr, bytes2Read)) < bytes2Read )
	{
	if (nBytes < 0)
	    goto error;
	bytes2Read -= nBytes;
	ptr += nBytes;
	}

    /* make sure all strxxx routines work... */

    fileContents[statbuf.st_size] = 0;

    /* Remove all comments from the file and convert everyting to lower cases */

    src = fileContents;
    dst = fileContents;
    inComment = 0;
    while (*src)
	{
	/* Comment recognition */

	if (*src == COMMENT_MARKER)	/* toggle comment in */
	    inComment = 1;
	if (*src == '\n')		/* toggle comment out */
	    inComment = 0;

	/* character modification */

	if (!inComment)
	    {
	    if (*src == COMMAND_SEPARATOR)
		{
		/* Do not add consequetive command separators */

		if ((dst != fileContents) &&
		    (*(dst - 1) != COMMAND_SEPARATOR))
		    {
		    *dst = COMMAND_SEPARATOR;
		    dst ++;
		    }
		}
	    else
		{
		*dst= *src;
		dst ++;
		}
	    }
	src ++;
	}

    *dst = '\0';

    /* Remove any leading "space" character */

    ptr = fileContents;

    while (*ptr)
    	{
	if (isspace ((int) *ptr))
	    ptr++;
	else
	    break;
	}

    /* Remove any ending "space" character */

    nBytes = strlen (ptr);

    for (;;)
    	{
	if (nBytes && isspace ((int) ptr [nBytes - 1]))
	    {
	    ptr[nBytes - 1] = EOS;
	    nBytes --;
	    }
	else
	    break;
	}

    if (strlen (ptr) != 0)
    	{
    	*pInitString = malloc (strlen (ptr) + 1);
	strcpy (*pInitString, ptr);
	}
    else
    	*pInitString = NULL;

    free (fileContents);
    return (0);
error:
    free (fileContents);
    return (-1);
    }

/*******************************************************************************
*
* main - The Wind River multi-protocol debug proxy
*
* wrproxy main entry point for all types of platform
*
* RETURNS: N/A
*
* ERRNO: N/A
*/

int main
    (
    int		argc,		/* number of arguments        */
    char *	argv[]		/* table of arguments         */
    )
    {
    char * initString = NULL;

#ifndef	WIN32
    /* Mask SIGPIPE signal */

    signal (SIGPIPE, SIG_IGN);
#endif	/* !WIN32 */

    /* Parse command line */

    if (cmdLineParse (argc, argv) == ERROR)
	return (ERROR);

    /* Parse startup script is one if specified */

    if (startupFileName != NULL)
    	{
	if (wrproxyStartupFileParse (startupFileName, &initString) == ERROR)
	    return (ERROR);
	}

    /* Register various supported backends */

    wrProxyUdpSockRegister ();
    wrProxyTcpSockRegister ();
#ifdef	LINUX
#ifndef WITHOUT_TIPC
    wrProxyTipcSockRegister ();
#endif /* !WITHOUT_TIPC */
#endif	/* LINUX */

    /* Start proxy server */

    /* TCP port number */
    if (verboseMode && hasControlPort)
	printf("Using TCP port number 0x%x (%u)\n",
			listeningTcpPort, listeningTcpPort);

    return (wrProxyLibInit (listeningTcpPort, initString, hasControlPort));
    }
